#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import json

import pytest
import responses

from datarobot import errors
from datarobot.models.secondary_dataset import SecondaryDataset, SecondaryDatasetConfigurations
from tests.utils import request_body_to_json


@pytest.fixture
def ids():
    return {
        "pid": "5a530498d5c1f302d6d17699",
        "graph_id": "5a530498d5c1f302d6d17111",
        "identifier_1": "test_dataset_1",
        "catalog_version_id_1": "5a530498d5c1f302d6d17121",
        "catalog_id_1": "5a530498d5c1f302d6d17115",
        "identifier_2": "test_dataset_2",
        "catalog_version_id_2": "6a530498d5c1f302d6d17121",
        "catalog_id_2": "6a530498d5c1f302d6d17115",
    }


@pytest.fixture
def secondary_datasets(ids):
    secondary_dataset_1 = SecondaryDataset(
        identifier=ids["identifier_1"],
        catalog_version_id=ids["catalog_version_id_1"],
        catalog_id=ids["catalog_id_1"],
        snapshot_policy="specified",
    )
    secondary_dataset_2 = SecondaryDataset(
        identifier=ids["identifier_2"],
        catalog_version_id=ids["catalog_version_id_2"],
        catalog_id=ids["catalog_id_2"],
        snapshot_policy="specified",
    )
    return [secondary_dataset_1, secondary_dataset_2]


@pytest.fixture
def secondary_dataset_configurations_response(ids):
    return {
        "projectId": ids["pid"],
        "secondaryDatasets": [
            {
                "snapshotPolicy": "specified",
                "identifier": ids["identifier_1"],
                "catalogVersionId": ids["catalog_version_id_1"],
                "catalogId": ids["catalog_id_1"],
            },
            {
                "snapshotPolicy": "specified",
                "identifier": ids["identifier_2"],
                "catalogVersionId": ids["catalog_version_id_2"],
                "catalogId": ids["catalog_id_2"],
            },
        ],
        "id": "5df109112ca582033ff44084",
        "name": "Testing",
    }


@responses.activate
def test_create_secondary_dataset_configurations(
    ids, secondary_datasets, secondary_dataset_configurations_response
):
    url = "https://host_name.com/projects/{}/secondaryDatasetsConfigurations/".format(ids["pid"])
    responses.add(
        responses.POST, url, status=201, body=json.dumps(secondary_dataset_configurations_response),
    )

    datasets = [dataset.to_dict() for dataset in secondary_datasets]
    result = SecondaryDatasetConfigurations.create(
        ids["pid"], secondary_datasets=datasets, name="Testing"
    )

    assert responses.calls[0].request.method == "POST"

    # verify request payload
    payload = request_body_to_json(responses.calls[0].request)
    all_secondary_datasets = payload["secondaryDatasets"]
    assert len(all_secondary_datasets) == 2

    # verify response
    assert isinstance(result, SecondaryDatasetConfigurations)
    assert result.project_id == ids["pid"]
    assert result.name == "Testing"
    assert len(result.secondary_datasets) == 2
    dataset = result.secondary_datasets[0]
    assert dataset.catalog_id == ids["catalog_id_1"]
    assert dataset.catalog_version_id == ids["catalog_version_id_1"]
    assert dataset.snapshot_policy == "specified"


@responses.activate
def test_failure_in_create_secondary_dataset_configurations__status_error_code(
    ids, secondary_datasets
):
    url = "https://host_name.com/projects/{}/secondaryDatasetsConfigurations/".format(ids["pid"])
    responses.add(responses.POST, url, status=404, body="")

    datasets = [dataset.to_dict() for dataset in secondary_datasets]
    with pytest.raises(errors.ClientError):
        SecondaryDatasetConfigurations.create(
            ids["pid"], secondary_datasets=datasets, name="Testing"
        )


@responses.activate
def test_secondary_dataset_configurations_retrieve(secondary_dataset_configurations_response):
    config_id = secondary_dataset_configurations_response["id"]
    pid = secondary_dataset_configurations_response["projectId"]
    url = "https://host_name.com/projects/{}/secondaryDatasetsConfigurations/{}/".format(
        pid, config_id
    )
    responses.add(responses.GET, url, json=secondary_dataset_configurations_response)

    secondary_dataset_config = SecondaryDatasetConfigurations(id=config_id, project_id=pid).get()

    assert responses.calls[0].request.method == "GET"
    assert secondary_dataset_config.id == config_id
    assert secondary_dataset_config.project_id == pid

    dataset = secondary_dataset_config.secondary_datasets[0]
    assert isinstance(dataset, SecondaryDataset)
    identifier = dataset.identifier
    expected_config = secondary_dataset_configurations_response["secondaryDatasets"]
    expected_dataset = [d for d in expected_config if d["identifier"] == identifier][0]
    assert dataset.catalog_id == expected_dataset["catalogId"]
    assert dataset.catalog_version_id == expected_dataset["catalogVersionId"]
    assert dataset.snapshot_policy == expected_dataset["snapshotPolicy"]


@responses.activate
def test_secondary_dataset_configurations_deletion(secondary_dataset_configurations_response):
    config_id = secondary_dataset_configurations_response["id"]
    pid = secondary_dataset_configurations_response["projectId"]
    url = "https://host_name.com/projects/{}/secondaryDatasetsConfigurations/{}/".format(
        pid, config_id
    )
    responses.add(responses.GET, url, json=secondary_dataset_configurations_response)
    responses.add(responses.DELETE, url)
    secondary_dataset_config = SecondaryDatasetConfigurations(id=config_id, project_id=pid).get()
    secondary_dataset_config.delete()
    assert responses.calls[1].request.method == responses.DELETE
    assert responses.calls[1].request.url == url


@responses.activate
def test_secondary_dataset_configurations_list(secondary_dataset_configurations_response):
    config_id = secondary_dataset_configurations_response["id"]
    pid = secondary_dataset_configurations_response["projectId"]
    config_list = {
        "data": [
            secondary_dataset_configurations_response,
            secondary_dataset_configurations_response,
        ]
    }
    url = "https://host_name.com/projects/{}/secondaryDatasetsConfigurations/".format(pid)
    responses.add(responses.GET, url, json=config_list)

    secondary_dataset_config = SecondaryDatasetConfigurations().list(project_id=pid)

    assert responses.calls[0].request.method == "GET"
    assert len(secondary_dataset_config) == 2
    config = secondary_dataset_config[0]
    assert config.id == config_id
    assert config.project_id == pid

    dataset = config.secondary_datasets[0]
    assert isinstance(dataset, SecondaryDataset)
    identifier = dataset.identifier
    expected_config = secondary_dataset_configurations_response["secondaryDatasets"]
    expected_dataset = [d for d in expected_config if d["identifier"] == identifier][0]
    assert dataset.catalog_id == expected_dataset["catalogId"]
    assert dataset.catalog_version_id == expected_dataset["catalogVersionId"]
    assert dataset.snapshot_policy == expected_dataset["snapshotPolicy"]


@responses.activate
def test_secondary_dataset_configurations(secondary_datasets):
    result = SecondaryDatasetConfigurations(
        id="c_id", project_id="p_id", config=[], secondary_datasets=secondary_datasets
    )
    assert isinstance(result, SecondaryDatasetConfigurations)
    result_config = result.to_dict()
    assert isinstance(result_config["secondary_datasets"], list)
    assert isinstance(result_config["secondary_datasets"][0], dict)
    assert isinstance(result_config["secondary_datasets"][1], dict)



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import pytest
import responses

import datarobot as dr
from datarobot.utils import parse_time


@pytest.fixture
def autodocs_endpoint(client):
    return "{}/{}".format(client.endpoint, dr.AutomatedDocument._path)


@pytest.fixture
def doc_id():
    return "5f6e3f01bcc52b0ef65bd013"


@pytest.fixture
def doc_url(doc_id):
    return "https://host_name.com/{}{}/".format(dr.AutomatedDocument._path, doc_id)


class TestAutomatedDocument:
    @responses.activate
    def test_list_available_document_types__success(self, client):
        url = "{}/automatedDocumentOptions/".format(client.endpoint)
        response_body = {"data": [{"document_type": "AUTOPILOT_SUMMARY", "locale": "EN_US"}]}
        responses.add(responses.GET, url, status=200, json=response_body)

        options = dr.AutomatedDocument.list_available_document_types()
        assert options == response_body

    @responses.activate
    def test_generate__success(self, autodocs_endpoint, doc_id, doc_url, client):
        document_type = "MODEL_COMPLIANCE"
        entity_id = "6f50cdb77cc4f8d1560c3ed5"
        output_format = "docx"
        locale = "EN_US"
        template_id = "50efc9db8aff6c81a374aeec"

        request_body = {
            "document_type": document_type,
            "entity_id": entity_id,
            "output_format": output_format,
            "locale": locale,
            "template_id": template_id,
        }

        status_url = "{}/status/17f182fa-6562-4cdc-9fe1-46ecbe769629/".format(client.endpoint)

        responses.add(
            responses.POST,
            autodocs_endpoint,
            json=request_body,
            status=202,
            adding_headers={"Location": status_url},
        )

        responses.add(responses.GET, status_url, status=303, adding_headers={"Location": doc_url})

        doc = dr.AutomatedDocument(
            document_type=document_type,
            entity_id=entity_id,
            output_format=output_format,
            locale=locale,
            template_id=template_id,
        )

        response = doc.generate()

        assert response.status_code == 202
        assert doc.id == doc_id
        assert doc.filepath is None
        assert doc.created_at is None

    @responses.activate
    def test_download__success(self, doc_url, doc_id):
        responses.add(
            responses.GET,
            doc_url,
            stream=True,
            status=200,
            content_type="application/docx",
            adding_headers={"Content-Disposition": "attachment;filename=report.docx"},
        )

        doc = dr.AutomatedDocument(id=doc_id)
        response = doc.download()

        assert response.status_code == 200
        assert set(response.headers) == {"Content-Disposition", "Content-Type"}
        assert doc.filepath == "report.docx"

    def test_download__no_doc_id__attr_error(self):
        with pytest.raises(AttributeError):
            dr.AutomatedDocument().download()

    @responses.activate
    def test_delete__success(self, doc_id, doc_url):
        responses.add(responses.DELETE, doc_url, status=204)

        doc = dr.AutomatedDocument(id=doc_id)
        response = doc.delete()

        assert response.status_code == 204

    def test_delete__no_doc_id__attr_error(self):
        with pytest.raises(AttributeError):
            dr.AutomatedDocument().delete()

    @responses.activate
    def test_list_generated_documents__success(self, autodocs_endpoint):

        items = [
            {
                "entity_id": "5e8c6a34d2427053ab4a39ed",
                "document_type": "MODEL_COMPLIANCE",
                "output_format": "docx",
                "locale": "EN_US",
                "template_id": "nul",
                "id": "60548987c297f6053d056f2e",
                "created_at": "2021-03-19 11:22:47.188000",
            },
            {
                "entity_id": "5bdc5f60d25eff701ad8fe91",
                "document_type": "AUTOPILOT_SUMMARY",
                "output_format": "html",
                "locale": "EN_US",
                "template_id": "513dce50edc297fc87a9db13ed6",
                "id": "60548969c297f6053d056f25",
                "created_at": "2020-10-29 10:27:50.866000",
            },
        ]

        json = {"previous": None, "next": None, "count": 2, "totalCount": 2, "data": items}

        responses.add(
            responses.GET, autodocs_endpoint, json=json, status=200,
        )

        [
            item.update({"created_at": parse_time(item["created_at"]), "filepath": None})
            for item in items
        ]

        docs = dr.AutomatedDocument.list_generated_documents()
        docs_for_compare = [doc.__dict__ for doc in docs]

        assert len(docs) == len(items)
        assert docs_for_compare == items



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from datetime import datetime
import json

from dateutil.tz import tzutc
import pytest
import responses

from datarobot import DatetimeModel
from tests.utils import request_body_to_json


@pytest.fixture
def accuracy_over_time_plots_metadata_response():
    return {
        "holdoutStatuses": {"training": "notCompleted", "validation": "completed"},
        "holdoutMetadata": {
            "training": {"startDate": None, "endDate": None},
            "validation": {
                "startDate": "2014-04-09T00:00:00.000000Z",
                "endDate": "2014-06-15T00:00:00.000000Z",
            },
        },
        "backtestMetadata": [
            {
                "training": {"startDate": None, "endDate": None},
                "validation": {"startDate": None, "endDate": None},
            },
            {
                "training": {"startDate": None, "endDate": None},
                "validation": {
                    "startDate": "2013-11-26T00:00:00.000000Z",
                    "endDate": "2014-02-01T00:00:00.000000Z",
                },
            },
        ],
        "forecastDistance": 1,
        "backtestStatuses": [
            {"training": "errored", "validation": "insufficientData"},
            {"training": "notCompleted", "validation": "completed"},
        ],
        "resolutions": ["days", "weeks", "months", "quarters", "years"],
    }


@pytest.fixture
def accuracy_over_time_plot_response():
    return {
        "startDate": "2014-02-01T00:00:00.000000Z",
        "statistics": {"durbinWatson": 0.5},
        "endDate": "2014-02-03T00:00:00.000000Z",
        "resolution": "days",
        "bins": [
            {
                "startDate": "2014-02-01T00:00:00.000000Z",
                "frequency": 10.0,
                "actual": 1.11,
                "predicted": 1.12,
                "endDate": "2014-02-02T00:00:00.000000Z",
            },
            {
                "startDate": "2014-02-02T00:00:00.000000Z",
                "frequency": 0.0,
                "actual": None,
                "predicted": None,
                "endDate": "2014-02-03T00:00:00.000000Z",
            },
        ],
        "calendarEvents": [
            {
                "date": "2014-02-17T00:00:00.000000Z",
                "seriesId": None,
                "name": "Washington's Birthday",
            }
        ],
    }


@pytest.fixture
def datetime_trend_plot_preview_response():
    return {
        "startDate": "2014-02-01T00:00:00.000000Z",
        "endDate": "2014-04-09T00:00:00.000000Z",
        "bins": [
            {
                "actual": 100,
                "predicted": 101,
                "startDate": "2014-02-01T00:00:00.000000Z",
                "endDate": "2014-02-02T00:00:00.000000Z",
            },
            {
                "actual": 100,
                "predicted": 101,
                "startDate": "2014-04-08T00:00:00.000000Z",
                "endDate": "2014-04-09T00:00:00.000000Z",
            },
        ],
    }


@pytest.fixture
def forecast_vs_actual_plots_metadata_response():
    return {
        "resolutions": ["days", "weeks", "months", "quarters", "years"],
        "backtestStatuses": [
            {
                "training": {"notCompleted": [1, 2, 3, 4, 5, 6, 7]},
                "validation": {"completed": [1], "errored": [2], "notCompleted": [3, 4, 5, 6, 7]},
            },
            {
                "training": {"notCompleted": [1, 2, 3, 4, 5, 6, 7]},
                "validation": {"inProgress": [1, 2, 3, 4, 5, 6, 7]},
            },
        ],
        "holdoutStatuses": {
            "training": {"notSupported": [1, 2, 3, 4, 5, 6, 7]},
            "validation": {"insufficientData": [1, 2, 3, 4, 5, 6, 7]},
        },
        "backtestMetadata": [
            {
                "training": {"startDate": None, "endDate": None},
                "validation": {
                    "startDate": "2014-02-01T00:00:00.000000Z",
                    "endDate": "2014-04-09T00:00:00.000000Z",
                },
            },
            {
                "training": {"startDate": None, "endDate": None},
                "validation": {"startDate": None, "endDate": None},
            },
        ],
        "holdoutMetadata": {
            "training": {"startDate": None, "endDate": None},
            "validation": {"startDate": None, "endDate": None},
        },
    }


@pytest.fixture
def forecast_vs_actual_plot_response():
    return {
        "startDate": "2014-02-01T00:00:00.000000Z",
        "endDate": "2014-04-09T00:00:00.000000Z",
        "resolution": "days",
        "forecastDistances": [1],
        "bins": [
            {
                "actual": 56363,
                "frequency": 10.0,
                "forecasts": [58441],
                "startDate": "2014-02-01T00:00:00.000000Z",
                "endDate": "2014-04-09T00:00:00.000000Z",
                "error": 15024,
                "normalizedError": 0.56,
            },
        ],
        "calendarEvents": [
            {
                "name": "Washington's Birthday",
                "date": "2014-02-17T00:00:00.000000Z",
                "seriesId": None,
            }
        ],
    }


@pytest.fixture
def anomaly_over_time_plots_metadata_response(accuracy_over_time_plots_metadata_response):
    response = accuracy_over_time_plots_metadata_response
    response.pop("forecastDistance")
    return response


@pytest.fixture
def anomaly_over_time_plot_response():
    return {
        "startDate": "2014-02-01T00:00:00.000000Z",
        "endDate": "2014-02-03T00:00:00.000000Z",
        "resolution": "days",
        "bins": [
            {
                "startDate": "2014-02-01T00:00:00.000000Z",
                "frequency": 10.0,
                "predicted": 1.12,
                "endDate": "2014-02-02T00:00:00.000000Z",
            },
            {
                "startDate": "2014-02-02T00:00:00.000000Z",
                "frequency": 0.0,
                "predicted": None,
                "endDate": "2014-02-03T00:00:00.000000Z",
            },
        ],
        "calendarEvents": [
            {
                "date": "2014-02-17T00:00:00.000000Z",
                "seriesId": None,
                "name": "Washington's Birthday",
            }
        ],
    }


@pytest.fixture
def anomaly_over_time_plot_preview_response():
    return {
        "startDate": "1970-01-01T10:30:01.000000Z",
        "endDate": "1970-01-01T11:30:01.000000Z",
        "predictionThreshold": 0.015,
        "bins": [
            {"startDate": "1970-01-01T10:33:01.000000Z", "endDate": "1970-01-01T10:36:01.000000Z"},
            {"startDate": "1970-01-01T10:36:01.000000Z", "endDate": "1970-01-01T10:39:01.000000Z"},
        ],
    }


@responses.activate
def test_get_accuracy_over_time_plots_metadata(
    accuracy_over_time_plots_metadata_response, unittest_endpoint, project_id, model_id
):
    responses.add(
        responses.GET,
        "{}/projects/{}/datetimeModels/{}/accuracyOverTimePlots/metadata/".format(
            unittest_endpoint, project_id, model_id
        ),
        status=200,
        content_type="application/json",
        body=json.dumps(accuracy_over_time_plots_metadata_response),
    )
    model = DatetimeModel(id=model_id, project_id=project_id)
    metadata = model.get_accuracy_over_time_plots_metadata(forecast_distance=1)
    assert metadata.project_id == project_id
    assert metadata.model_id == model_id
    assert metadata.resolutions == accuracy_over_time_plots_metadata_response["resolutions"]
    assert (
        metadata.forecast_distance == accuracy_over_time_plots_metadata_response["forecastDistance"]
    )
    assert metadata.backtest_statuses[0] == {
        "training": "errored",
        "validation": "insufficientData",
    }
    assert metadata.backtest_statuses[1] == {"training": "notCompleted", "validation": "completed"}
    assert metadata.backtest_metadata[1]["training"] == {"end_date": None, "start_date": None}
    assert metadata.backtest_metadata[1]["validation"] == {
        "end_date": datetime(2014, 2, 1, tzinfo=tzutc()),
        "start_date": datetime(2013, 11, 26, tzinfo=tzutc()),
    }

    assert metadata.holdout_statuses == {"training": "notCompleted", "validation": "completed"}
    assert metadata.holdout_metadata["validation"] == {
        "end_date": datetime(2014, 6, 15, tzinfo=tzutc()),
        "start_date": datetime(2014, 4, 9, tzinfo=tzutc()),
    }

    assert metadata._get_status(backtest=1) == "completed"
    assert metadata._get_status(backtest=1, source="training") == "notCompleted"
    assert metadata._get_status(backtest="holdout", source="training") == "notCompleted"
    assert metadata._get_status(backtest="holdout", source="validation") == "completed"

    assert metadata._get_status(backtest="holdout!") is None
    assert metadata._get_status(backtest=10) is None
    assert metadata._get_status(source="validation!") is None


@responses.activate
def test_get_accuracy_over_time_plot(
    accuracy_over_time_plot_response, unittest_endpoint, project_id, model_id
):
    responses.add(
        responses.GET,
        "{}/projects/{}/datetimeModels/{}/accuracyOverTimePlots/".format(
            unittest_endpoint, project_id, model_id
        ),
        status=200,
        content_type="application/json",
        body=json.dumps(accuracy_over_time_plot_response),
    )
    model = DatetimeModel(id=model_id, project_id=project_id)
    plot = model.get_accuracy_over_time_plot(
        start_date=datetime(2014, 2, 2), end_date=datetime(2014, 2, 3), max_wait=None
    )
    assert plot.project_id == project_id
    assert plot.model_id == model_id
    assert plot.resolution == accuracy_over_time_plot_response["resolution"]
    assert plot.bins[0] == {
        "actual": 1.11,
        "end_date": datetime(2014, 2, 2, tzinfo=tzutc()),
        "frequency": 10,
        "predicted": 1.12,
        "start_date": datetime(2014, 2, 1, tzinfo=tzutc()),
    }
    assert plot.calendar_events[0] == {
        "date": datetime(2014, 2, 17, tzinfo=tzutc()),
        "name": "Washington's Birthday",
        "series_id": None,
    }
    assert plot.statistics["durbin_watson"] == 0.5


@responses.activate
@pytest.mark.parametrize("plot", ["accuracyOverTime", "forecastVsActual"])
def test_get_datetime_trend_plot_preview(
    datetime_trend_plot_preview_response, unittest_endpoint, project_id, model_id, plot
):
    responses.add(
        responses.GET,
        "{}/projects/{}/datetimeModels/{}/{}Plots/preview/".format(
            unittest_endpoint, project_id, model_id, plot
        ),
        status=200,
        content_type="application/json",
        body=json.dumps(datetime_trend_plot_preview_response),
    )
    model = DatetimeModel(id=model_id, project_id=project_id)
    plot = (
        model.get_accuracy_over_time_plot_preview(max_wait=None)
        if "accuracy" in plot
        else model.get_forecast_vs_actual_plot_preview(max_wait=None)
    )
    assert plot.project_id == project_id
    assert plot.model_id == model_id
    assert plot.start_date == datetime(2014, 2, 1, tzinfo=tzutc())
    assert plot.bins[0] == {
        "actual": 100,
        "end_date": datetime(2014, 2, 2, tzinfo=tzutc()),
        "predicted": 101,
        "start_date": datetime(2014, 2, 1, tzinfo=tzutc()),
    }


@responses.activate
def test_get_forecast_vs_actual_plots_metadata(
    forecast_vs_actual_plots_metadata_response, unittest_endpoint, project_id, model_id
):
    responses.add(
        responses.GET,
        "{}/projects/{}/datetimeModels/{}/forecastVsActualPlots/metadata/".format(
            unittest_endpoint, project_id, model_id
        ),
        status=200,
        content_type="application/json",
        body=json.dumps(forecast_vs_actual_plots_metadata_response),
    )
    model = DatetimeModel(id=model_id, project_id=project_id)
    metadata = model.get_forecast_vs_actual_plots_metadata()
    assert metadata.project_id == project_id
    assert metadata.model_id == model_id
    assert metadata.resolutions == forecast_vs_actual_plots_metadata_response["resolutions"]
    assert metadata.backtest_statuses == [
        {
            "training": {"notCompleted": [1, 2, 3, 4, 5, 6, 7]},
            "validation": {"completed": [1], "notCompleted": [3, 4, 5, 6, 7], "errored": [2]},
        },
        {
            "training": {"notCompleted": [1, 2, 3, 4, 5, 6, 7]},
            "validation": {"inProgress": [1, 2, 3, 4, 5, 6, 7]},
        },
    ]
    assert metadata.backtest_metadata[0]["validation"] == {
        "end_date": datetime(2014, 4, 9, tzinfo=tzutc()),
        "start_date": datetime(2014, 2, 1, tzinfo=tzutc()),
    }
    assert metadata.holdout_statuses == {
        "training": {"notSupported": [1, 2, 3, 4, 5, 6, 7]},
        "validation": {"insufficientData": [1, 2, 3, 4, 5, 6, 7]},
    }

    assert metadata._get_status(backtest=1) == {"inProgress": [1, 2, 3, 4, 5, 6, 7]}
    assert metadata._get_status(backtest=1, source="training") == {
        "notCompleted": [1, 2, 3, 4, 5, 6, 7]
    }
    assert metadata._get_status(backtest="holdout", source="training") == {
        "notSupported": [1, 2, 3, 4, 5, 6, 7]
    }
    assert metadata._get_status(backtest="holdout", source="validation") == {
        "insufficientData": [1, 2, 3, 4, 5, 6, 7]
    }


@responses.activate
def test_get_forecast_vs_actual_plot(
    forecast_vs_actual_plot_response, unittest_endpoint, project_id, model_id
):
    responses.add(
        responses.GET,
        "{}/projects/{}/datetimeModels/{}/forecastVsActualPlots/".format(
            unittest_endpoint, project_id, model_id
        ),
        status=200,
        content_type="application/json",
        body=json.dumps(forecast_vs_actual_plot_response),
    )
    model = DatetimeModel(id=model_id, project_id=project_id)
    plot = model.get_forecast_vs_actual_plot(
        start_date=datetime(2014, 2, 2), end_date=datetime(2014, 2, 3), max_wait=None
    )
    assert plot.project_id == project_id
    assert plot.model_id == model_id
    assert plot.resolution == forecast_vs_actual_plot_response["resolution"]
    assert plot.forecast_distances == forecast_vs_actual_plot_response["forecastDistances"]
    assert plot.bins[0] == {
        "actual": 56363,
        "frequency": 10.0,
        "forecasts": [58441],
        "start_date": datetime(2014, 2, 1, tzinfo=tzutc()),
        "end_date": datetime(2014, 4, 9, tzinfo=tzutc()),
        "error": 15024,
        "normalized_error": 0.56,
    }
    assert plot.calendar_events[0] == {
        "date": datetime(2014, 2, 17, tzinfo=tzutc()),
        "name": "Washington's Birthday",
        "series_id": None,
    }


@responses.activate
def test_get_anomaly_over_time_plots_metadata(
    anomaly_over_time_plots_metadata_response, unittest_endpoint, project_id, model_id
):
    responses.add(
        responses.GET,
        "{}/projects/{}/datetimeModels/{}/anomalyOverTimePlots/metadata/".format(
            unittest_endpoint, project_id, model_id
        ),
        status=200,
        content_type="application/json",
        body=json.dumps(anomaly_over_time_plots_metadata_response),
    )
    model = DatetimeModel(id=model_id, project_id=project_id)
    metadata = model.get_anomaly_over_time_plots_metadata()
    assert metadata.project_id == project_id
    assert metadata.model_id == model_id
    assert metadata.resolutions == anomaly_over_time_plots_metadata_response["resolutions"]
    assert metadata.backtest_statuses[0] == {
        "training": "errored",
        "validation": "insufficientData",
    }
    assert metadata.backtest_statuses[1] == {"training": "notCompleted", "validation": "completed"}
    assert metadata.backtest_metadata[1]["training"] == {"end_date": None, "start_date": None}
    assert metadata.backtest_metadata[1]["validation"] == {
        "end_date": datetime(2014, 2, 1, tzinfo=tzutc()),
        "start_date": datetime(2013, 11, 26, tzinfo=tzutc()),
    }


@responses.activate
def test_get_anomaly_over_time_plot(
    anomaly_over_time_plot_response, unittest_endpoint, project_id, model_id
):
    responses.add(
        responses.GET,
        "{}/projects/{}/datetimeModels/{}/anomalyOverTimePlots/".format(
            unittest_endpoint, project_id, model_id
        ),
        status=200,
        content_type="application/json",
        body=json.dumps(anomaly_over_time_plot_response),
    )
    model = DatetimeModel(id=model_id, project_id=project_id)
    plot = model.get_anomaly_over_time_plot(
        start_date=datetime(2014, 2, 2), end_date=datetime(2014, 2, 3), max_wait=None
    )
    assert plot.project_id == project_id
    assert plot.model_id == model_id
    assert plot.resolution == anomaly_over_time_plot_response["resolution"]
    assert plot.bins[0] == {
        "end_date": datetime(2014, 2, 2, tzinfo=tzutc()),
        "frequency": 10,
        "predicted": 1.12,
        "start_date": datetime(2014, 2, 1, tzinfo=tzutc()),
    }
    assert plot.calendar_events[0] == {
        "date": datetime(2014, 2, 17, tzinfo=tzutc()),
        "name": "Washington's Birthday",
        "series_id": None,
    }


@responses.activate
def test_get_anomaly_over_time_plot_preview(
    anomaly_over_time_plot_preview_response, unittest_endpoint, project_id, model_id
):
    responses.add(
        responses.GET,
        "{}/projects/{}/datetimeModels/{}/anomalyOverTimePlots/preview/".format(
            unittest_endpoint, project_id, model_id
        ),
        status=200,
        content_type="application/json",
        body=json.dumps(anomaly_over_time_plot_preview_response),
    )
    model = DatetimeModel(id=model_id, project_id=project_id)
    plot = model.get_anomaly_over_time_plot_preview(max_wait=None)

    assert plot.prediction_threshold == 0.015
    assert plot.project_id == project_id
    assert plot.model_id == model_id
    assert plot.start_date == datetime(1970, 1, 1, 10, 30, 1, tzinfo=tzutc())
    assert plot.bins[0] == {
        "end_date": datetime(1970, 1, 1, 10, 36, 1, tzinfo=tzutc()),
        "start_date": datetime(1970, 1, 1, 10, 33, 1, tzinfo=tzutc()),
    }


def _add_compute_route_to_responses(unittest_endpoint, project_id, model_id, responses):
    compute_url = "{}/projects/{}/datetimeModels/{}/datetimeTrendPlots/".format(
        unittest_endpoint, project_id, model_id
    )
    job_url = "{}/projects/{}/jobs/1/".format(unittest_endpoint, project_id)

    responses.add(responses.POST, compute_url, status=202, headers={"Location": job_url})
    responses.add(
        responses.GET,
        job_url,
        status=303,
        content_type="application/json",
        body=json.dumps(
            {
                "status": "COMPLETED",
                "model_id": model_id,
                "is_blocked": False,
                "url": job_url,
                "job_type": "datetime_trend_plots",
                "project_id": project_id,
                "id": 1,
            }
        ),
        headers={"Location": "{}result/".format(job_url)},
    )


@responses.activate
def test_accuracy_over_time_plots_auto_compute_before_retrieval(
    unittest_endpoint, project_id, model_id, accuracy_over_time_plots_metadata_response,
):
    def reset_responses():
        responses.reset()
        _add_compute_route_to_responses(unittest_endpoint, project_id, model_id, responses)
        responses.add(
            responses.GET,
            "{}/projects/{}/datetimeModels/{}/accuracyOverTimePlots/metadata/".format(
                unittest_endpoint, project_id, model_id
            ),
            status=200,
            content_type="application/json",
            body=json.dumps(accuracy_over_time_plots_metadata_response),
        )

    model = DatetimeModel(id=model_id, project_id=project_id)

    # Backtest 1, Training, FD 10: Status is not completed. Run compute job
    reset_responses()
    model._compute_accuracy_over_time_plot_if_not_computed(
        backtest=1, source="training", forecast_distance=10, max_wait=100
    )
    assert request_body_to_json(responses.calls[1].request) == {
        "backtest": 1,
        "source": "training",
        "forecastDistanceStart": 10,
        "forecastDistanceEnd": 10,
    }

    # Backtest 1, Training, FD 10: Status is completed. Do not run compute job
    reset_responses()
    model._compute_accuracy_over_time_plot_if_not_computed(
        backtest=0, source="training", forecast_distance=10, max_wait=100
    )
    assert len(responses.calls) == 1 and responses.calls[0].request.method == responses.GET


@responses.activate
def test_forecast_vs_actual_plots_auto_compute_before_retrieval(
    unittest_endpoint, project_id, model_id, forecast_vs_actual_plots_metadata_response,
):
    def reset_responses():
        responses.reset()
        _add_compute_route_to_responses(unittest_endpoint, project_id, model_id, responses)
        responses.add(
            responses.GET,
            "{}/projects/{}/datetimeModels/{}/forecastVsActualPlots/metadata/".format(
                unittest_endpoint, project_id, model_id
            ),
            status=200,
            content_type="application/json",
            body=json.dumps(forecast_vs_actual_plots_metadata_response),
        )

    model = DatetimeModel(id=model_id, project_id=project_id)

    # Backtest 0, Validation: FD 2 is errored, but FD 3 and FD 4 are not completed. Run compute job
    reset_responses()
    model._compute_forecast_vs_actual_plot_if_not_computed(
        backtest=0,
        source="validation",
        forecast_distance_start=2,
        forecast_distance_end=4,
        max_wait=100,
    )
    assert request_body_to_json(responses.calls[1].request) == {
        "backtest": 0,
        "source": "validation",
        "forecastDistanceStart": 2,
        "forecastDistanceEnd": 4,
    }

    # Backtest 0, Validation: FD 1 is completed, FD 2 is errored. Do not run compute job
    reset_responses()
    model._compute_forecast_vs_actual_plot_if_not_computed(
        backtest=0,
        source="validation",
        forecast_distance_start=1,
        forecast_distance_end=2,
        max_wait=100,
    )
    assert len(responses.calls) == 1 and responses.calls[0].request.method == responses.GET

    # Backtest 1, Validation: All FDs are in progress. Do not run compute job
    reset_responses()
    model._compute_forecast_vs_actual_plot_if_not_computed(
        backtest=1,
        source="validation",
        forecast_distance_start=None,
        forecast_distance_end=None,
        max_wait=100,
    )
    assert len(responses.calls) == 1 and responses.calls[0].request.method == responses.GET

    # Backtest 0, Validation: FD 1 is completed, FD 2 is errored, but FD 3-4 are not completed.
    # Run compute job
    reset_responses()
    model._compute_forecast_vs_actual_plot_if_not_computed(
        backtest=0,
        source="validation",
        forecast_distance_start=None,
        forecast_distance_end=4,
        max_wait=100,
    )
    assert request_body_to_json(responses.calls[1].request) == {
        "backtest": 0,
        "source": "validation",
        "forecastDistanceEnd": 4,
    }

    # Backtest 0, Validation: FD 3-7 are not completed. Run compute job
    reset_responses()
    model._compute_forecast_vs_actual_plot_if_not_computed(
        backtest=0,
        source="validation",
        forecast_distance_start=3,
        forecast_distance_end=None,
        max_wait=100,
    )
    assert request_body_to_json(responses.calls[1].request) == {
        "backtest": 0,
        "source": "validation",
        "forecastDistanceStart": 3,
    }

    # Backtest 0, Validation: FD 1 is completed, FD 2 is errored, FD 3-7 are not completed.
    # Run compute job
    reset_responses()
    model._compute_forecast_vs_actual_plot_if_not_computed(
        backtest=0,
        source="validation",
        forecast_distance_start=None,
        forecast_distance_end=None,
        max_wait=100,
    )

    assert request_body_to_json(responses.calls[1].request) == {
        "backtest": 0,
        "source": "validation",
    }


@responses.activate
def test_anomaly_over_time_plots_auto_compute_before_retrieval(
    unittest_endpoint, project_id, model_id, anomaly_over_time_plots_metadata_response,
):
    def reset_responses():
        responses.reset()
        _add_compute_route_to_responses(unittest_endpoint, project_id, model_id, responses)
        responses.add(
            responses.GET,
            "{}/projects/{}/datetimeModels/{}/anomalyOverTimePlots/metadata/".format(
                unittest_endpoint, project_id, model_id
            ),
            status=200,
            content_type="application/json",
            body=json.dumps(anomaly_over_time_plots_metadata_response),
        )

    model = DatetimeModel(id=model_id, project_id=project_id)

    # Backtest 1, Training: Status is not completed. Run compute job
    reset_responses()
    model._compute_anomaly_over_time_plot_if_not_computed(
        backtest=1, source="training", max_wait=100
    )
    assert request_body_to_json(responses.calls[1].request) == {
        "backtest": 1,
        "source": "training",
    }

    # Backtest 1, Training: Status is completed. Do not run compute job
    reset_responses()
    model._compute_anomaly_over_time_plot_if_not_computed(
        backtest=0, source="training", max_wait=100
    )
    assert len(responses.calls) == 1 and responses.calls[0].request.method == responses.GET



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import pytest
from trafaret import DataError

from datarobot.models.custom_task_version_dependency_build import CustomTaskVersionDependencyBuild
from tests.model.utils import assert_custom_task_version_dependency_build


@pytest.mark.parametrize(
    "missing_field", ["customTaskId", "customTaskVersionId", "buildStart", "buildStatus"],
)
def test_from_server_data__fail_with_missing_field(
    missing_field, custom_task_version_dependency_build_server_data
):
    custom_task_version_dependency_build_server_data.pop(missing_field)
    with pytest.raises(DataError):
        CustomTaskVersionDependencyBuild.from_server_data(
            custom_task_version_dependency_build_server_data
        )


@pytest.mark.parametrize(
    "field_null_not_allowed", ["customTaskId", "customTaskVersionId", "buildStart", "buildStatus"],
)
def test_from_server_data__fail__field_with_blank_values(
    field_null_not_allowed, custom_task_version_dependency_build_server_data,
):
    custom_task_version_dependency_build_server_data.update({field_null_not_allowed: None})
    with pytest.raises(DataError):
        CustomTaskVersionDependencyBuild.from_server_data(
            custom_task_version_dependency_build_server_data
        )


@pytest.mark.parametrize(
    "field_null_allowed", ["buildEnd", "buildLogLocation"],
)
def test_from_server_data__succeed__field_with_blank_values(
    field_null_allowed, custom_task_version_dependency_build_server_data,
):
    custom_task_version_dependency_build_server_data.update({field_null_allowed: None})
    dependency_build = CustomTaskVersionDependencyBuild.from_server_data(
        custom_task_version_dependency_build_server_data
    )
    assert_custom_task_version_dependency_build(
        dependency_build, custom_task_version_dependency_build_server_data
    )


def test_from_server_data__succeed_with_field_naming_conversion(
    custom_task_version_dependency_build_server_data,
):
    dependency_build = CustomTaskVersionDependencyBuild.from_server_data(
        custom_task_version_dependency_build_server_data
    )
    assert_custom_task_version_dependency_build(
        dependency_build, custom_task_version_dependency_build_server_data
    )


def test_from_server_data__succeed_ignoring_extra_fields(
    custom_task_version_dependency_build_server_data,
):
    custom_task_version_dependency_build_server_data["extra_fields"] = "aaa"
    dependency_build = CustomTaskVersionDependencyBuild.from_server_data(
        custom_task_version_dependency_build_server_data
    )
    assert_custom_task_version_dependency_build(
        dependency_build, custom_task_version_dependency_build_server_data
    )



# coding: utf-8
#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import json

import responses

from datarobot.models.model import ClusteringModel


@responses.activate
def test_get_clustering_model_with_new_fields(
    project_url, project_id, model_id, clustering_models, project_clustering_data
):
    responses.add(responses.GET, project_url, body=json.dumps(project_clustering_data))

    for model_data in clustering_models:

        model_id = model_data["id"]

        responses.add(
            responses.GET,
            "{}models/{}/".format(project_url, model_id),
            body=json.dumps(model_data),
        )
        model = ClusteringModel.get(project_id, model_id)

        assert model.id == model_data["id"]
        assert model.model_type == model_data["modelType"]
        assert model.project_id == model_data["projectId"]
        # new fields
        assert model.n_clusters == model_data["nClusters"]
        assert (
            model.is_n_clusters_dynamically_determined
            == model_data["isNClustersDynamicallyDetermined"]
        )
        assert model.has_empty_clusters == model_data["hasEmptyClusters"]


def test_clustering_model_future_proof(clustering_model_gmm_server_data):
    data_with_future_keys = dict(clustering_model_gmm_server_data, new_key="new key")
    ClusteringModel.from_server_data(data_with_future_keys)



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from datetime import datetime
import json

from dateutil.tz import tzutc
import pytest
import responses
from trafaret import DataError

from datarobot._experimental.models.segmentation import SegmentationTask

_project_id = "projectId"
_target = "target"
_model_package_id = "modelPackageId"
_current_time = "2020-01-01T01:00:01.123456Z"
_base_url = "https://host_name.com"
_create_url = "{}/projects/{}/segmentationTasks/"
_results_url = "{}/projects/{}/segmentationTasks/results/{}/"
_get_url = "{}/projects/{}/segmentationTasks/{}/"

_valid_segmentation_task_id = "segmentationTaskId"
_invalid_segmentation_task_id = "badSegmentationTaskId"
_valid_segementation_task_url = _get_url.format(_base_url, _project_id, _valid_segmentation_task_id)
_invalid_segementation_task_url = _get_url.format(
    _base_url, _project_id, _invalid_segmentation_task_id
)


@pytest.fixture
def segmentation_record():
    return {
        "projectId": _project_id,
        "segmentationTaskId": _valid_segmentation_task_id,
        "name": "testSegmentor",
        "type": "testType",
        "created": _current_time,
        "segmentsCount": 1,
        "segments": ["user_segment_1"],
        "metadata": {
            "useMultiseriesIdColumns": True,
            "useUserDefinedSegmentColumns": True,
            "useTimeSeries": True,
            "useAutomatedSegmentation": False,
        },
        "data": {
            "multiseriesIdColumns": ["series_id"],
            "userDefinedSegmentIdColumns": ["segment"],
            "datetimePartitionColumn": "date",
        },
    }


@pytest.fixture
def segmentation_failed_record():
    return {
        "name": "badTestSegmentor",
        "parameters": {"parameter": "value"},
        "message": "errorMessage",
    }


@pytest.fixture
def segmentation_results_response(segmentation_record, segmentation_failed_record):
    return {
        "numberOfJobs": 2,
        "completedJobs": [
            {
                "name": segmentation_record["name"],
                "segmentationTaskId": segmentation_record["segmentationTaskId"],
                "segmentsCount": segmentation_record["segmentsCount"],
                "url": _get_url.format(
                    _base_url,
                    segmentation_record["projectId"],
                    segmentation_record["segmentationTaskId"],
                ),
            }
        ],
        "failedJobs": [segmentation_failed_record],
    }


@pytest.fixture
def segmentation_results_response_json(segmentation_results_response):
    return json.dumps(segmentation_results_response)


@pytest.fixture
def segmentation_list_response(segmentation_record):
    return {"count": 1, "next": None, "previous": None, "data": [segmentation_record]}


@pytest.fixture
def segmentation_task_get_json(segmentation_record):
    return json.dumps(segmentation_record)


@pytest.fixture
def invalid_json():
    return json.dumps({"message": "Not Found"})


@pytest.fixture
def segmentation_task_list_json(segmentation_list_response):
    return json.dumps(segmentation_list_response)


@pytest.fixture
def automated_segmentation_record():
    return {
        "projectId": _project_id,
        "segmentationTaskId": _valid_segmentation_task_id,
        "name": "testAutomatedSegmentor",
        "type": "testType",
        "created": _current_time,
        "segmentsCount": 3,
        "segments": ["segment_1", "segment_2", "segment_3"],
        "metadata": {
            "useMultiseriesIdColumns": True,
            "useUserDefinedSegmentColumns": False,
            "useTimeSeries": True,
            "useAutomatedSegmentation": True,
        },
        "data": {
            "multiseriesIdColumns": ["series_id"],
            "datetimePartitionColumn": "date",
            "modelPackageId": _model_package_id,
        },
    }


@pytest.fixture
def automated_segmentation_task_get_json(automated_segmentation_record):
    return json.dumps(automated_segmentation_record)


@pytest.fixture
def automated_segmentation_results_response(
    automated_segmentation_record, segmentation_failed_record
):
    return {
        "numberOfJobs": 1,
        "completedJobs": [
            {
                "name": automated_segmentation_record["name"],
                "segmentationTaskId": automated_segmentation_record["segmentationTaskId"],
                "segmentsCount": automated_segmentation_record["segmentsCount"],
                "url": _get_url.format(
                    _base_url,
                    automated_segmentation_record["projectId"],
                    automated_segmentation_record["segmentationTaskId"],
                ),
            }
        ],
        "failedJobs": [segmentation_failed_record],
    }


@pytest.fixture
def automated_segmentation_results_response_json(automated_segmentation_results_response):
    return json.dumps(automated_segmentation_results_response)


def test_time_series_must_have_date(
    automated_segmentation_results_response_json, automated_segmentation_task_get_json
):
    with pytest.raises(
        ValueError, match="A datetime_partition_column value must be specified for time series."
    ):
        SegmentationTask.create(
            _project_id, _target, use_time_series=True,
        )


def test_time_series_must_have_multiseries_id(
    automated_segmentation_results_response_json, automated_segmentation_task_get_json
):
    with pytest.raises(
        ValueError, match="A multiseries_id_columns value must be specified for time series."
    ):
        SegmentationTask.create(
            _project_id, _target, use_time_series=True, datetime_partition_column="date",
        )


def test_time_series_multiseries_id_must_be_list_or_tuple(
    automated_segmentation_results_response_json, automated_segmentation_task_get_json
):
    with pytest.raises(
        ValueError, match="Expected list of str for multiseries_id_columns, got: True"
    ):
        SegmentationTask.create(
            _project_id,
            _target,
            use_time_series=True,
            datetime_partition_column="date",
            multiseries_id_columns=True,
        )


def test_segmentation_must_be_user_defined_or_automatic(
    automated_segmentation_results_response_json, automated_segmentation_task_get_json
):
    with pytest.raises(
        ValueError,
        match="Either a user_defined_segment_id_columns value or a model_package_id must "
        "be defined to create a new segmentation task.",
    ):
        SegmentationTask.create(
            _project_id,
            _target,
            use_time_series=True,
            datetime_partition_column="date",
            multiseries_id_columns=["series_id"],
        )


@responses.activate
def test_use_automated_segmentation_task_creation(
    automated_segmentation_results_response_json, automated_segmentation_task_get_json
):
    responses.add(
        responses.POST,
        _create_url.format(_base_url, _project_id),
        body="",
        status=202,
        adding_headers={"Location": "{}/status/some-status-id/".format(_base_url)},
        content_type="application_json",
    )
    responses.add(
        responses.GET,
        "{}/status/some-status-id/".format(_base_url),
        body="",
        status=303,
        adding_headers={"Location": _results_url.format(_base_url, _project_id, "resultsId")},
        content_type="application/json",
    )
    responses.add(
        responses.GET,
        _results_url.format(_base_url, _project_id, "resultsId"),
        body=automated_segmentation_results_response_json,
        status=200,
        content_type="application/json",
    )
    responses.add(
        responses.GET,
        _valid_segementation_task_url,
        body=automated_segmentation_task_get_json,
        status=200,
        content_type="application/json",
    )
    segmentation_task_results = SegmentationTask.create(
        _project_id,
        _target,
        use_time_series=True,
        datetime_partition_column="date",
        multiseries_id_columns=["series_id"],
        model_package_id=_model_package_id,
    )
    assert segmentation_task_results["numberOfJobs"] == 1
    assert len(segmentation_task_results["completedJobs"]) == 1
    segmentation_task = segmentation_task_results["completedJobs"][0]
    assert isinstance(segmentation_task, SegmentationTask)
    assert segmentation_task.id == _valid_segmentation_task_id
    assert segmentation_task.project_id == _project_id
    assert segmentation_task.created == datetime(2020, 1, 1, 1, 0, 1, 123456, tzinfo=tzutc())
    assert segmentation_task.name == "testAutomatedSegmentor"
    assert segmentation_task.type == "testType"
    assert segmentation_task.segments_count == 3
    assert segmentation_task.segments == ["segment_1", "segment_2", "segment_3"]
    assert segmentation_task.metadata == {
        "use_multiseries_id_columns": True,
        "use_user_defined_segment_columns": False,
        "use_time_series": True,
        "use_automated_segmentation": True,
    }
    assert segmentation_task.data == {
        "multiseries_id_columns": ["series_id"],
        "datetime_partition_column": "date",
        "model_package_id": "modelPackageId",
    }


@responses.activate
def test_segmentation_create(
    segmentation_results_response_json, segmentation_task_get_json, segmentation_failed_record
):
    responses.add(
        responses.POST,
        _create_url.format(_base_url, _project_id),
        body="",
        status=202,
        adding_headers={"Location": "{}/status/some-status-id/".format(_base_url)},
        content_type="application_json",
    )
    responses.add(
        responses.GET,
        "{}/status/some-status-id/".format(_base_url),
        body="",
        status=303,
        adding_headers={"Location": _results_url.format(_base_url, _project_id, "resultsId")},
        content_type="application/json",
    )
    responses.add(
        responses.GET,
        _results_url.format(_base_url, _project_id, "resultsId"),
        body=segmentation_results_response_json,
        status=200,
        content_type="application/json",
    )
    responses.add(
        responses.GET,
        _valid_segementation_task_url,
        body=segmentation_task_get_json,
        status=200,
        content_type="application/json",
    )
    segmentation_task_results = SegmentationTask.create(
        _project_id,
        _target,
        use_time_series=True,
        datetime_partition_column="date",
        multiseries_id_columns=["series_id"],
        user_defined_segment_id_columns=["segment_id"],
    )
    assert segmentation_task_results["numberOfJobs"] == 2
    assert len(segmentation_task_results["completedJobs"]) == 1
    segmentation_task = segmentation_task_results["completedJobs"][0]

    assert isinstance(segmentation_task, SegmentationTask)
    assert segmentation_task.id == _valid_segmentation_task_id
    assert segmentation_task.project_id == _project_id
    assert segmentation_task.created == datetime(2020, 1, 1, 1, 0, 1, 123456, tzinfo=tzutc())
    assert segmentation_task.name == "testSegmentor"
    assert segmentation_task.type == "testType"
    assert segmentation_task.segments_count == 1
    assert segmentation_task.segments == ["user_segment_1"]
    assert segmentation_task.metadata == {
        "use_multiseries_id_columns": True,
        "use_user_defined_segment_columns": True,
        "use_time_series": True,
        "use_automated_segmentation": False,
    }
    assert segmentation_task.data == {
        "multiseries_id_columns": ["series_id"],
        "user_defined_segment_id_columns": ["segment"],
        "datetime_partition_column": "date",
    }

    assert segmentation_task_results["failedJobs"] == [segmentation_failed_record]


@responses.activate
def test_segmentation_list(segmentation_task_list_json):
    responses.add(
        responses.GET,
        _create_url.format(_base_url, _project_id),
        body=segmentation_task_list_json,
        status=200,
        content_type="application/json",
    )
    segmentation_tasks = SegmentationTask.list(_project_id)
    assert len(segmentation_tasks) == 1
    segmentation_task = segmentation_tasks[0]

    assert isinstance(segmentation_task, SegmentationTask)
    assert segmentation_task.id == _valid_segmentation_task_id
    assert segmentation_task.project_id == _project_id
    assert segmentation_task.created == datetime(2020, 1, 1, 1, 0, 1, 123456, tzinfo=tzutc())
    assert segmentation_task.name == "testSegmentor"
    assert segmentation_task.type == "testType"
    assert segmentation_task.segments_count == 1
    assert segmentation_task.segments == ["user_segment_1"]
    assert segmentation_task.metadata == {
        "use_multiseries_id_columns": True,
        "use_user_defined_segment_columns": True,
        "use_time_series": True,
        "use_automated_segmentation": False,
    }
    assert segmentation_task.data == {
        "multiseries_id_columns": ["series_id"],
        "user_defined_segment_id_columns": ["segment"],
        "datetime_partition_column": "date",
    }


@responses.activate
def test_segmentation_get_valid(segmentation_task_get_json):
    responses.add(
        responses.GET,
        _valid_segementation_task_url,
        body=segmentation_task_get_json,
        status=200,
        content_type="application/json",
    )
    segmentation_task = SegmentationTask.get(_project_id, _valid_segmentation_task_id)

    assert isinstance(segmentation_task, SegmentationTask)
    assert segmentation_task.id == _valid_segmentation_task_id
    assert segmentation_task.project_id == _project_id
    assert segmentation_task.created == datetime(2020, 1, 1, 1, 0, 1, 123456, tzinfo=tzutc())
    assert segmentation_task.name == "testSegmentor"
    assert segmentation_task.type == "testType"
    assert segmentation_task.segments_count == 1
    assert segmentation_task.segments == ["user_segment_1"]
    assert segmentation_task.metadata == {
        "use_multiseries_id_columns": True,
        "use_user_defined_segment_columns": True,
        "use_time_series": True,
        "use_automated_segmentation": False,
    }
    assert segmentation_task.data == {
        "multiseries_id_columns": ["series_id"],
        "user_defined_segment_id_columns": ["segment"],
        "datetime_partition_column": "date",
    }


@responses.activate
def test_segmentation_task_invalid_get(invalid_json):
    responses.add(responses.GET, _invalid_segementation_task_url, body=invalid_json)
    with pytest.raises(DataError):
        SegmentationTask.get(_project_id, _invalid_segmentation_task_id)



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import mock
import pytest

from datarobot.client import get_client
from tests.utils import SDKTestcase


class TestBuildRequestWithFile(SDKTestcase):
    def test_file_doesnt_exist(self):
        client = get_client()
        with self.assertRaises(ValueError):
            client.build_request_with_file(
                "PUT", "http://host-name/", "meh.csv", file_path="meh.csv"
            )

    def test_non_string_content(self):
        client = get_client()
        with self.assertRaises(AssertionError):
            client.build_request_with_file("PUT", "http://host-name/", "meh.csv", content=1234)


def test_join_endpoint(client):
    """Ensure no // in generated endpoints."""
    with mock.patch.object(client, "endpoint", "http://host_name.local/api/v2"):
        assert (
            client._join_endpoint("asdf/endpoint/")
            == "http://host_name.local/api/v2/asdf/endpoint/"
        )
        with pytest.raises(ValueError):
            # /absolute/paths are rejected
            client._join_endpoint("/asdf/endpoint/")
    with mock.patch.object(client, "endpoint", "http://host_name.local/api/v2/"):
        assert (
            client._join_endpoint("asdf/endpoint/")
            == "http://host_name.local/api/v2/asdf/endpoint/"
        )
    with mock.patch.object(client, "endpoint", "http://host_name.local"):
        assert client._join_endpoint("asdf/endpoint/") == "http://host_name.local/asdf/endpoint/"
    with mock.patch.object(client, "endpoint", "http://host_name.local/"):
        assert client._join_endpoint("asdf/endpoint/") == "http://host_name.local/asdf/endpoint/"


def test_strip_endpoint(client):
    url = "https://host_name.com/the/url/I/want/"
    stripped = client.strip_endpoint(url)
    assert stripped == "the/url/I/want/"


def test_strip_endpoint_with_training_slash(client):
    client.endpoint = "https://host_name.com/"
    url = "https://host_name.com/the/url/I/want/"
    stripped = client.strip_endpoint(url)
    assert stripped == "the/url/I/want/"


def test_strip_endpoint_invalid_url(client):
    url = "http://localhost/api/v2/wrong/prefix/on/this/url"
    with pytest.raises(ValueError):
        client.strip_endpoint(url)



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import pytest

from datarobot import (
    BacktestSpecification,
    DatetimeModel,
    DatetimePartitioning,
    DatetimePartitioningSpecification,
    enums,
    FeatureSettings,
    Periodicity,
    Project,
)
from datarobot.utils import parse_time


@pytest.fixture
def holdout_start_date():
    return parse_time("2015-12-10T19:00:00.000000Z")


@pytest.fixture
def holdout_end_date():
    return parse_time("2015-12-24T19:00:00.000000Z")


@pytest.fixture
def datetime_partition_without_holdout_server_data(project_id):
    return {
        "primaryTrainingEndDate": None,
        "holdoutRowCount": 0,
        "backtests": [
            {
                "index": 0,
                "validationRowCount": 108,
                "primaryTrainingDuration": "P0Y0M48DT0H5M36S",
                "primaryTrainingEndDate": "2014-01-07T22:44:23.000000Z",
                "availableTrainingStartDate": "2013-11-20T22:38:47.000000Z",
                "primaryTrainingStartDate": "2013-11-20T22:38:47.000000Z",
                "validationEndDate": "2014-01-12T22:44:23.000000Z",
                "availableTrainingDuration": "P0Y0M48DT0H5M36S",
                "availableTrainingRowCount": 4168,
                "gapEndDate": "2014-01-07T22:44:23.000000Z",
                "validationDuration": "P5D",
                "gapStartDate": "2014-01-07T22:44:23.000000Z",
                "availableTrainingEndDate": "2014-01-07T22:44:23.000000Z",
                "primaryTrainingRowCount": 4168,
                "validationStartDate": "2014-01-07T22:44:23.000000Z",
                "totalRowCount": 4276,
                "gapRowCount": 0,
                "gapDuration": "P0Y0M0D",
            }
        ],
        "availableTrainingDuration": "P0Y0M53DT0H5M36S",
        "availableTrainingRowCount": 4276,
        "gapEndDate": None,
        "partitioningWarnings": [],
        "primaryTrainingRowCount": 0,
        "holdoutStartDate": None,
        "datetimePartitionColumn": "Timestamp",
        "dateFormat": "%Y-%m-%d %H:%M:%S",
        "useTimeSeries": False,
        "defaultToAPriori": False,
        "defaultToKnownInAdvance": False,
        "featureDerivationWindowStart": None,
        "featureDerivationWindowEnd": None,
        "forecastWindowStart": None,
        "forecastWindowEnd": None,
        "numberOfBacktests": 1,
        "availableTrainingStartDate": "2013-11-20T22:38:47.000000Z",
        "primaryTrainingStartDate": None,
        "holdoutEndDate": None,
        "gapStartDate": None,
        "totalRowCount": 4276,
        "gapRowCount": 0,
        "primaryTrainingDuration": "P0Y0M0D",
        "holdoutDuration": "P0Y0M0D",
        "gapDuration": "P0Y0M0D",
        "autopilotDataSelectionMethod": "duration",
        "validationDuration": "P5D",
        "projectId": project_id,
        "availableTrainingEndDate": "2014-01-12T22:44:23.000000Z",
        "numberOfKnownInAdvanceFeatures": 1,
        "numberOfDoNotDeriveFeatures": 1,
        "calendarId": None,
        "calendarName": None,
    }


@pytest.fixture
def datetime_partition_server_data(project_id, holdout_start_date, holdout_end_date):
    return {
        "primaryTrainingEndDate": "2015-12-10T19:00:00.000000Z",
        "primaryTrainingDuration": "P0Y1M0D",
        "backtests": [
            {
                "primaryTrainingEndDate": "2015-11-26T19:00:00.000000Z",
                "primaryTrainingDuration": "P0Y1M0D",
                "index": 0,
                "availableTrainingStartDate": "2015-06-08T20:00:00.000000Z",
                "primaryTrainingStartDate": "2015-10-26T19:00:00.000000Z",
                "validationEndDate": "2015-12-10T19:00:00.000000Z",
                "availableTrainingDuration": "P0Y0M170DT23H0M0S",
                "validationDuration": "P14D",
                "gapStartDate": "2015-11-26T19:00:00.000000Z",
                "availableTrainingEndDate": "2015-11-26T19:00:00.000000Z",
                "validationStartDate": "2015-11-26T19:00:00.000000Z",
                "gapEndDate": "2015-11-26T19:00:00.000000Z",
                "gapDuration": "P0Y0M0D",
            }
        ],
        "datetimePartitionColumn": "dates",
        "useTimeSeries": False,
        "defaultToAPriori": False,
        "defaultToKnownInAdvance": False,
        "featureDerivationWindowStart": None,
        "featureDerivationWindowEnd": None,
        "forecastWindowStart": None,
        "forecastWindowEnd": None,
        "dateFormat": "%Y-%m-%d %H:%M:%S",
        "projectId": project_id,
        "availableTrainingStartDate": "2015-06-08T20:00:00.000000Z",
        "primaryTrainingStartDate": "2015-11-10T19:00:00.000000Z",
        "holdoutEndDate": holdout_end_date.isoformat(),
        "validationDuration": "P14D",
        "availableTrainingDuration": "P0Y0M184DT23H0M0S",
        "numberOfBacktests": 1,
        "gapStartDate": "2015-12-10T19:00:00.000000Z",
        "availableTrainingEndDate": "2015-12-10T19:00:00.000000Z",
        "holdoutStartDate": holdout_start_date.isoformat(),
        "gapEndDate": "2015-12-10T19:00:00.000000Z",
        "holdoutDuration": "P14D",
        "gapDuration": "P0Y0M0D",
        "autopilotDataSelectionMethod": "rowCount",
        "numberOfKnownInAdvanceFeatures": 1,
        "numberOfDoNotDeriveFeatures": 1,
        "calendarId": None,
        "calendarName": None,
        "modelSplits": 2,
        "allowPartialHistoryTimeSeriesPredictions": False,
    }


@pytest.fixture
def datetime_partition_server_data_feature_settings(datetime_partition_server_data):
    settings = dict(datetime_partition_server_data)
    settings.update(
        {
            "featureSettings": [
                {"featureName": "timestamp", "knownInAdvance": True, "aPriori": True},
                {"featureName": "target", "knownInAdvance": True, "aPriori": True},
                {"featureName": "series_id", "knownInAdvance": False, "aPriori": False},
                {"featureName": "input3", "knownInAdvance": False, "aPriori": False},
                {"featureName": "input4", "knownInAdvance": True, "aPriori": True},
                {"featureName": "input5", "doNotDerive": True, "knownInAdvance": False},
                {"featureName": "input6", "doNotDerive": False, "knownInAdvance": True},
            ],
        }
    )
    return settings


@pytest.fixture
def datetime_partition_server_data_extra_time_series_controls(datetime_partition_server_data):
    settings = dict(datetime_partition_server_data)
    settings.update(
        {
            "periodicities": [
                {"timeSteps": 10, "timeUnit": "HOUR"},
                {"timeSteps": 600, "timeUnit": "MINUTE"},
                {"timeSteps": 7, "timeUnit": "DAY"},
            ],
            "treatAsExponential": enums.TREAT_AS_EXPONENTIAL.ALWAYS,
            "differencingMethod": enums.DIFFERENCING_METHOD.SIMPLE,
        }
    )
    return settings


@pytest.fixture
def datetime_partition_time_series_seq_based_windows_server_data(
    datetime_partition_time_series_server_data,
):
    settings = dict(datetime_partition_time_series_server_data)
    settings.update(
        {"windowsBasisUnit": "ROW", "periodicities": [{"timeSteps": 10, "timeUnit": "ROW"}]}
    )
    return settings


@pytest.fixture
def datetime_partition_milliseconds_server_data(datetime_partition_time_series_server_data):
    settings = dict(datetime_partition_time_series_server_data)
    settings.update(
        {
            "windowsBasisUnit": "MILLISECOND",
            "periodicities": [{"timeSteps": 500, "timeUnit": "MILLISECOND"}],
        }
    )
    return settings


@pytest.fixture
def datetime_partition_irregular_time_series_server_data(datetime_partition_server_data):
    settings = dict(datetime_partition_server_data)
    time_series_settings = {
        "useTimeSeries": True,
        "featureDerivationWindowStart": -9,
        "featureDerivationWindowEnd": -5,
        "forecastWindowStart": 5,
        "forecastWindowEnd": 9,
        "validationDuration": None,
    }
    settings.update(time_series_settings)
    return settings


@pytest.fixture
def datetime_partition_time_series_server_data(datetime_partition_server_data):
    settings = dict(datetime_partition_server_data)
    time_series_settings = {
        "useTimeSeries": True,
        "defaultToKnownInAdvance": True,
        "defaultToDoNotDerive": True,
        "featureDerivationWindowStart": -9,
        "featureDerivationWindowEnd": -5,
        "forecastWindowStart": 5,
        "forecastWindowEnd": 9,
    }
    settings.update(time_series_settings)
    return settings


@pytest.fixture
def datetime_partition_after_target_server_data(project_id, holdout_start_date, holdout_end_date):
    return {
        "primaryTrainingEndDate": "2015-12-10T19:00:00.000000Z",
        "primaryTrainingDuration": "P0Y1M0D",
        "primaryTrainingRowCount": 100,
        "backtests": [
            {
                "primaryTrainingEndDate": "2015-11-26T19:00:00.000000Z",
                "primaryTrainingDuration": "P0Y1M0D",
                "primaryTrainingRowCount": 100,
                "index": 0,
                "availableTrainingStartDate": "2015-06-08T20:00:00.000000Z",
                "primaryTrainingStartDate": "2015-10-26T19:00:00.000000Z",
                "validationEndDate": "2015-12-10T19:00:00.000000Z",
                "availableTrainingDuration": "P0Y0M170DT23H0M0S",
                "availableTrainingRowCount": 100,
                "validationDuration": "P14D",
                "validationRowCount": 100,
                "gapStartDate": "2015-11-26T19:00:00.000000Z",
                "availableTrainingEndDate": "2015-11-26T19:00:00.000000Z",
                "validationStartDate": "2015-11-26T19:00:00.000000Z",
                "gapEndDate": "2015-11-26T19:00:00.000000Z",
                "gapDuration": "P0Y0M0D",
                "gapRowCount": 100,
                "totalRowCount": 100,
            }
        ],
        "datetimePartitionColumn": "dates",
        "dateFormat": "%Y-%m-%d %H:%M:%S",
        "useTimeSeries": False,
        "defaultToAPriori": False,
        "defaultToKnownInAdvance": False,
        "featureDerivationWindowStart": None,
        "featureDerivationWindowEnd": None,
        "forecastWindowStart": None,
        "forecastWindowEnd": None,
        "projectId": project_id,
        "availableTrainingStartDate": "2015-06-08T20:00:00.000000Z",
        "primaryTrainingStartDate": "2015-11-10T19:00:00.000000Z",
        "holdoutEndDate": holdout_end_date.isoformat(),
        "validationDuration": "P14D",
        "availableTrainingDuration": "P0Y0M184DT23H0M0S",
        "availableTrainingRowCount": 100,
        "numberOfBacktests": 1,
        "gapStartDate": "2015-12-10T19:00:00.000000Z",
        "availableTrainingEndDate": "2015-12-10T19:00:00.000000Z",
        "holdoutStartDate": holdout_start_date.isoformat(),
        "gapEndDate": "2015-12-10T19:00:00.000000Z",
        "holdoutDuration": "P14D",
        "holdoutRowCount": 100,
        "gapDuration": "P0Y0M0D",
        "gapRowCount": 100,
        "totalRowCount": 100,
        "autopilotDataSelectionMethod": "rowCount",
        "numberOfKnownInAdvanceFeatures": 1,
        "numberOfDoNotDeriveFeatures": 1,
        "calendarId": "some-id",
        "calendarName": "some name",
        "modelSplits": 2,
        "allowPartialTimeSeriesPredictions": False,
    }


@pytest.fixture
def datetime_partition_after_target_server_data_feature_settings(
    datetime_partition_after_target_server_data,
):
    settings = dict(datetime_partition_after_target_server_data)
    settings.update(
        {
            "featureSettings": [
                {"featureName": "timestamp", "knownInAdvance": True, "aPriori": True},
                {"featureName": "target", "knownInAdvance": True, "aPriori": True},
                {"featureName": "series_id", "knownInAdvance": True, "aPriori": True},
                {"featureName": "input3", "knownInAdvance": True, "aPriori": True},
                {"featureName": "input4", "knownInAdvance": True, "aPriori": True},
            ],
        }
    )
    return settings


@pytest.fixture
def datetime_partition(datetime_partition_server_data):
    return DatetimePartitioning.from_server_data(datetime_partition_server_data)


@pytest.fixture
def datetime_partition_spec(datetime_partition_server_data, holdout_start_date):
    backtest_data = datetime_partition_server_data["backtests"][0]
    return DatetimePartitioningSpecification(
        datetime_partition_server_data["datetimePartitionColumn"],
        autopilot_data_selection_method=enums.DATETIME_AUTOPILOT_DATA_SELECTION_METHOD.ROW_COUNT,
        validation_duration=datetime_partition_server_data["validationDuration"],
        holdout_start_date=holdout_start_date,
        holdout_duration=datetime_partition_server_data["holdoutDuration"],
        gap_duration=datetime_partition_server_data["gapDuration"],
        number_of_backtests=datetime_partition_server_data["numberOfBacktests"],
        backtests=[
            BacktestSpecification(
                backtest_data["index"],
                backtest_data["gapDuration"],
                parse_time(backtest_data["validationStartDate"]),
                backtest_data["validationDuration"],
            )
        ],
        model_splits=datetime_partition_server_data["modelSplits"],
    )


@pytest.fixture
def datetime_partition_spec_feature_settings(
    datetime_partition_server_data_feature_settings, holdout_start_date
):
    backtest_data = datetime_partition_server_data_feature_settings["backtests"][0]
    feature_settings_data = datetime_partition_server_data_feature_settings["featureSettings"]
    return DatetimePartitioningSpecification(
        datetime_partition_server_data_feature_settings["datetimePartitionColumn"],
        autopilot_data_selection_method=enums.DATETIME_AUTOPILOT_DATA_SELECTION_METHOD.ROW_COUNT,
        validation_duration=datetime_partition_server_data_feature_settings["validationDuration"],
        holdout_start_date=holdout_start_date,
        holdout_duration=datetime_partition_server_data_feature_settings["holdoutDuration"],
        gap_duration=datetime_partition_server_data_feature_settings["gapDuration"],
        number_of_backtests=datetime_partition_server_data_feature_settings["numberOfBacktests"],
        backtests=[
            BacktestSpecification(
                backtest_data["index"],
                backtest_data["gapDuration"],
                parse_time(backtest_data["validationStartDate"]),
                backtest_data["validationDuration"],
            )
        ],
        feature_settings=[
            FeatureSettings(
                fs["featureName"], fs.get("knownInAdvance", None), fs.get("doNotDerive", None)
            )
            for fs in feature_settings_data
        ],
    )


@pytest.fixture
def datetime_partition_spec_bad_feature_settings(
    datetime_partition_server_data_feature_settings, holdout_start_date
):
    backtest_data = datetime_partition_server_data_feature_settings["backtests"][0]
    feature_settings_data = datetime_partition_server_data_feature_settings["featureSettings"]
    return DatetimePartitioningSpecification(
        datetime_partition_server_data_feature_settings["datetimePartitionColumn"],
        autopilot_data_selection_method=enums.DATETIME_AUTOPILOT_DATA_SELECTION_METHOD.ROW_COUNT,
        validation_duration=datetime_partition_server_data_feature_settings["validationDuration"],
        holdout_start_date=holdout_start_date,
        holdout_duration=datetime_partition_server_data_feature_settings["holdoutDuration"],
        gap_duration=datetime_partition_server_data_feature_settings["gapDuration"],
        number_of_backtests=datetime_partition_server_data_feature_settings["numberOfBacktests"],
        backtests=[
            BacktestSpecification(
                backtest_data["index"],
                backtest_data["gapDuration"],
                parse_time(backtest_data["validationStartDate"]),
                backtest_data["validationDuration"],
            )
        ],
        feature_settings=[FeatureSettings(fs["featureName"], None) for fs in feature_settings_data],
    )


@pytest.fixture
def datetime_partition_spec_extra_time_series_controls(
    datetime_partition_server_data_extra_time_series_controls, holdout_start_date
):
    server_data = datetime_partition_server_data_extra_time_series_controls
    backtest_data = server_data["backtests"][0]
    return DatetimePartitioningSpecification(
        server_data["datetimePartitionColumn"],
        autopilot_data_selection_method=enums.DATETIME_AUTOPILOT_DATA_SELECTION_METHOD.ROW_COUNT,
        validation_duration=server_data["validationDuration"],
        holdout_start_date=holdout_start_date,
        holdout_duration=server_data["holdoutDuration"],
        gap_duration=server_data["gapDuration"],
        number_of_backtests=server_data["numberOfBacktests"],
        backtests=[
            BacktestSpecification(
                backtest_data["index"],
                backtest_data["gapDuration"],
                parse_time(backtest_data["validationStartDate"]),
                backtest_data["validationDuration"],
            )
        ],
        treat_as_exponential=server_data["treatAsExponential"],
        differencing_method=server_data["differencingMethod"],
        periodicities=[
            Periodicity(p["timeSteps"], p["timeUnit"]) for p in server_data["periodicities"]
        ],
    )


@pytest.fixture
def irregular_time_series_partition_spec(
    datetime_partition_irregular_time_series_server_data, holdout_start_date
):
    server_data = datetime_partition_irregular_time_series_server_data
    backtest_data = server_data["backtests"][0]
    return DatetimePartitioningSpecification(
        server_data["datetimePartitionColumn"],
        autopilot_data_selection_method=enums.DATETIME_AUTOPILOT_DATA_SELECTION_METHOD.ROW_COUNT,
        validation_duration=None,
        holdout_start_date=holdout_start_date,
        holdout_duration=server_data["holdoutDuration"],
        gap_duration=server_data["gapDuration"],
        number_of_backtests=server_data["numberOfBacktests"],
        backtests=[
            BacktestSpecification(
                backtest_data["index"],
                backtest_data["gapDuration"],
                parse_time(backtest_data["validationStartDate"]),
                backtest_data["validationDuration"],
            )
        ],
        use_time_series=server_data["useTimeSeries"],
        feature_derivation_window_start=server_data["featureDerivationWindowStart"],
        feature_derivation_window_end=server_data["featureDerivationWindowEnd"],
        forecast_window_start=server_data["forecastWindowStart"],
        forecast_window_end=server_data["forecastWindowEnd"],
    )


@pytest.fixture
def time_series_partition_spec(datetime_partition_time_series_server_data, holdout_start_date):
    backtest_data = datetime_partition_time_series_server_data["backtests"][0]
    fdw_start = datetime_partition_time_series_server_data["featureDerivationWindowStart"]
    fdw_end = datetime_partition_time_series_server_data["featureDerivationWindowEnd"]
    return DatetimePartitioningSpecification(
        datetime_partition_time_series_server_data["datetimePartitionColumn"],
        autopilot_data_selection_method=enums.DATETIME_AUTOPILOT_DATA_SELECTION_METHOD.ROW_COUNT,
        validation_duration=datetime_partition_time_series_server_data["validationDuration"],
        holdout_start_date=holdout_start_date,
        holdout_duration=datetime_partition_time_series_server_data["holdoutDuration"],
        gap_duration=datetime_partition_time_series_server_data["gapDuration"],
        number_of_backtests=datetime_partition_time_series_server_data["numberOfBacktests"],
        backtests=[
            BacktestSpecification(
                backtest_data["index"],
                backtest_data["gapDuration"],
                parse_time(backtest_data["validationStartDate"]),
                backtest_data["validationDuration"],
            )
        ],
        use_time_series=datetime_partition_time_series_server_data["useTimeSeries"],
        default_to_known_in_advance=datetime_partition_time_series_server_data[
            "defaultToKnownInAdvance"
        ],
        default_to_do_not_derive=datetime_partition_time_series_server_data["defaultToDoNotDerive"],
        feature_derivation_window_start=fdw_start,
        feature_derivation_window_end=fdw_end,
        forecast_window_start=datetime_partition_time_series_server_data["forecastWindowStart"],
        forecast_window_end=datetime_partition_time_series_server_data["forecastWindowEnd"],
    )


@pytest.fixture
def datetime_project_server_data(project_id):
    return {
        "id": project_id,
        "projectName": "Untitled Project",
        "fileName": "special_date_kickcars.csv",
        "stage": "modeling",
        "autopilotMode": 0,
        "created": "2017-01-05T17:13:51.257353Z",
        "target": "y",
        "metric": "LogLoss",
        "partition": {
            "datetimeCol": None,
            "cvMethod": "datetime",
            "datetimePartitionColumn": "Purch\xc3\xa4Date",
            "validationPct": None,
            "reps": None,
            "cvHoldoutLevel": None,
            "holdoutLevel": None,
            "userPartitionCol": None,
            "validationType": "TVH",
            "trainingLevel": None,
            "partitionKeyCols": None,
            "holdoutPct": None,
            "validationLevel": None,
        },
        "recommender": {
            "recommenderItemId": None,
            "isRecommender": None,
            "recommenderUserId": None,
        },
        "advancedOptions": {
            "responseCap": False,
            "downsampledMinorityRows": None,
            "downsampledMajorityRows": None,
            "blueprintThreshold": 3,
            "seed": None,
            "weights": None,
            "smartDownsampled": False,
            "majorityDownsamplingRate": None,
        },
        "positiveClass": 1,
        "maxTrainPct": 78.125,
        "holdoutUnlocked": False,
        "targetType": "Binary",
    }


@pytest.fixture
def datetime_project(datetime_project_server_data):
    return Project.from_server_data(datetime_project_server_data)


@pytest.fixture
def datetime_model_server_data(project_id):
    return {
        "featurelistId": "586ef743ccf94e7d7310c288",
        "processes": [
            "Ordinal encoding of categorical variables",
            "Missing Values Imputed",
            "Gradient Boosted Trees Classifier",
        ],
        "featurelistName": "Informative Features",
        "backtests": [
            {
                "index": 0,
                "score": 0.7,
                "status": "COMPLETED",
                "training_start_date": "2015-10-26T19:00:00.000000Z",
                "training_duration": "P0Y1M0D",
                "training_row_count": 880,
                "training_end_date": "2015-11-26T19:00:00.000000Z",
            },
            {
                "index": 1,
                "score": 0.4,
                "status": "COMPLETED",
                "training_start_date": "2015-10-26T19:00:00.000000Z",
                "training_duration": "P0Y1M0D",
                "training_row_count": 880,
                "training_end_date": "2015-11-26T19:00:00.000000Z",
            },
            {
                "index": 1,
                "score": None,
                "status": "BOUNDARIES_EXCEEDED",
                "training_start_date": None,
                "training_duration": None,
                "training_row_count": None,
                "training_end_date": None,
            },
        ],
        "modelType": "Gradient Boosted Trees Classifier",
        "modelCategory": "model",
        "projectId": project_id,
        "dataSelectionMethod": "rowCount",
        "samplePct": None,
        "holdoutStatus": "COMPLETED",
        "trainingStartDate": None,
        "metrics": {
            "AUC": {
                "backtesting": None,
                "holdout": None,
                "backtestingScores": [0.3, 0.23, None],
                "crossValidation": None,
                "validation": 0.75862,
            },
            "Rate@Top5%": {
                "backtesting": None,
                "holdout": None,
                "backtestingScores": [1.0, 2.5, None],
                "crossValidation": None,
                "validation": 0.75,
            },
            "Rate@TopTenth%": {
                "backtesting": None,
                "holdout": None,
                "backtestingScores": [1.223, 1.23, None],
                "crossValidation": None,
                "validation": 1,
            },
            "RMSE": {
                "backtesting": None,
                "holdout": None,
                "backtestingScores": [1.5, 0.2, None],
                "crossValidation": None,
                "validation": 0.3721,
            },
            "LogLoss": {
                "backtesting": None,
                "holdout": None,
                "backtestingScores": [0.7, 0.4, None],
                "crossValidation": None,
                "validation": 0.46308,
            },
            "FVE Binomial": {
                "backtesting": None,
                "holdout": None,
                "backtestingScores": [0.9, 0.8, None],
                "crossValidation": None,
                "validation": 0.1381,
            },
            "Gini Norm": {
                "backtesting": None,
                "holdout": None,
                "backtestingScores": [0.4, 0.3, None],
                "crossValidation": None,
                "validation": 0.51724,
            },
            "Rate@Top10%": {
                "backtesting": None,
                "holdout": None,
                "backtestingScores": [0.5, 0.4, None],
                "crossValidation": None,
                "validation": 0.84615,
            },
        },
        "trainingDuration": None,
        "trainingRowCount": 880,
        "holdoutScore": None,
        "trainingInfo": {
            "predictionTrainingDuration": "P0Y1M0D",
            "holdoutTrainingStartDate": "2015-10-26T19:00:00.000000Z",
            "predictionTrainingStartDate": "2015-10-26T19:00:00.000000Z",
            "holdoutTrainingDuration": "P0Y1M0D",
            "predictionTrainingRowCount": 880,
            "holdoutTrainingEndDate": "2015-12-10T19:00:00.000000Z",
            "predictionTrainingEndDate": "2015-12-10T19:00:00.000000Z",
            "holdoutTrainingRowCount": 880,
        },
        "isFrozen": False,
        "blueprintId": "588d02115164b77809d1223820c8933f",
        "timeWindowSamplePct": None,
        "trainingEndDate": None,
        "id": "586ef75cccf94e7da010c294",
        "effective_feature_derivation_window_start": -120,
        "effective_feature_derivation_window_end": 0,
        "forecast_window_start": 10,
        "forecast_window_end": 10,
        "windows_basis_unit": "MINUTE",
        "supportsMonotonicConstraints": True,
        "monotonicIncreasingFeaturelistId": "5ae04c0d962d7410683073cb",
        "monotonicDecreasingFeaturelistId": "5ae04c0f962d7410683073cc",
    }


@pytest.fixture
def datetime_start_end_model_server_data(datetime_model_server_data):
    final = dict(datetime_model_server_data)
    final["dataSelectionMethod"] = "selectedDateRange"
    final["trainingRowCount"] = None
    final["trainingStartDate"] = "2015-10-26T19:00:00.000000Z"
    final["trainingEndDate"] = "2015-12-10T19:00:00.000000Z"
    final["timeWindowSamplePct"] = 50
    return final


@pytest.fixture
def datetime_model(datetime_model_server_data):
    return DatetimeModel.from_server_data(datetime_model_server_data)


@pytest.fixture
def datetime_model_job_server_data(project_id):
    body = {
        "status": "inprogress",
        "processes": [
            "One-Hot Encoding",
            "Bernoulli Naive Bayes classifier (scikit-learn)",
            "Missing Values Imputed",
            "Gaussian Naive Bayes classifier (scikit-learn)",
            "Naive Bayes combiner classifier",
            "Calibrate predictions",
        ],
        "projectId": project_id,
        "samplePct": None,
        "modelType": "Gradient Boosted Trees Classifier",
        "featurelistId": "586ef743ccf94e7d7310c288",
        "modelCategory": "model",
        "blueprintId": "588d02115164b77809d1223820c8933f",
        "isBlocked": False,
        "id": "1",
    }
    return body



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import mock
import responses
import six

from datarobot import DataDriver


@responses.activate
def test_list(drivers_endpoint, drivers_list_server_resp):
    responses.add(responses.GET, drivers_endpoint, json=drivers_list_server_resp)
    drivers = DataDriver.list()
    for driver, server_payload in zip(drivers, drivers_list_server_resp["data"]):
        assert driver.id == server_payload["id"]
        assert driver.class_name == server_payload["className"]
        assert driver.base_names == server_payload["baseNames"]
        assert driver.canonical_name == server_payload["canonicalName"]
        assert driver.creator == server_payload["creator"]


@responses.activate
def test_get(drivers_endpoint, postgresql_driver_server_resp):
    driver_id = postgresql_driver_server_resp["id"]
    responses.add(
        responses.GET,
        "{}{}/".format(drivers_endpoint, driver_id),
        json=postgresql_driver_server_resp,
    )

    driver = DataDriver.get(driver_id)
    assert driver.id == postgresql_driver_server_resp["id"]
    assert driver.class_name == postgresql_driver_server_resp["className"]
    assert driver.base_names == postgresql_driver_server_resp["baseNames"]
    assert driver.canonical_name == postgresql_driver_server_resp["canonicalName"]
    assert driver.creator == postgresql_driver_server_resp["creator"]


@responses.activate
def test_future_proof(drivers_endpoint, postgresql_driver_server_resp):
    driver_id = postgresql_driver_server_resp["id"]
    resp = dict(postgresql_driver_server_resp, new_key="new")
    responses.add(responses.GET, "{}{}/".format(drivers_endpoint, driver_id), json=resp)

    driver = DataDriver.get(driver_id)
    assert driver.id == driver_id


@responses.activate
def test_create(
    drivers_endpoint,
    driver_libraries_endpoint,
    driver_library_upload_server_resp,
    postgresql_driver_server_resp,
):
    responses.add(responses.POST, driver_libraries_endpoint, json=driver_library_upload_server_resp)
    responses.add(responses.POST, drivers_endpoint, json=postgresql_driver_server_resp)

    with mock.patch("os.path.exists"):
        with mock.patch("datarobot.rest.open", create=True) as open_mock:
            open_mock.return_value = six.StringIO("thra\ntata\nrata")
            driver = DataDriver.create(
                class_name=postgresql_driver_server_resp["className"],
                canonical_name=postgresql_driver_server_resp["canonicalName"],
                files=[
                    "/tmp/{}".format(name) for name in postgresql_driver_server_resp["baseNames"]
                ],
            )
    assert driver.id == postgresql_driver_server_resp["id"]
    assert driver.class_name == postgresql_driver_server_resp["className"]
    assert driver.base_names == postgresql_driver_server_resp["baseNames"]
    assert driver.canonical_name == postgresql_driver_server_resp["canonicalName"]
    assert driver.creator == postgresql_driver_server_resp["creator"]


@responses.activate
def test_create_local_multiple_libraries(
    drivers_endpoint,
    driver_libraries_endpoint,
    driver_library_upload_server_resp,
    postgresql_driver_server_resp,
):
    postgresql_driver_server_resp["baseNames"] += postgresql_driver_server_resp["baseNames"]
    responses.add(responses.POST, driver_libraries_endpoint, json=driver_library_upload_server_resp)
    responses.add(responses.POST, drivers_endpoint, json=postgresql_driver_server_resp)

    with mock.patch("os.path.exists"):
        with mock.patch("datarobot.rest.open", create=True) as open_mock:
            open_mock.return_value = six.StringIO("thra\ntata\nrata")
            # send with multiple libraries
            driver = DataDriver.create(
                class_name=postgresql_driver_server_resp["className"],
                canonical_name=postgresql_driver_server_resp["canonicalName"],
                files=[
                    "/tmp/{}".format(name) for name in postgresql_driver_server_resp["baseNames"]
                ],
            )
    assert driver.id == postgresql_driver_server_resp["id"]
    assert driver.class_name == postgresql_driver_server_resp["className"]
    assert driver.base_names == postgresql_driver_server_resp["baseNames"]
    assert driver.canonical_name == postgresql_driver_server_resp["canonicalName"]
    assert driver.creator == postgresql_driver_server_resp["creator"]


@responses.activate
def test_update(drivers_endpoint, postgresql_driver_server_resp):
    driver_id = postgresql_driver_server_resp["id"]
    responses.add(
        responses.GET,
        "{}{}/".format(drivers_endpoint, driver_id),
        json=postgresql_driver_server_resp,
    )

    updated_postgresql_driver_server_resp = dict(
        postgresql_driver_server_resp, className="NewClassName", canonicalName="NewCanonicalName"
    )

    responses.add(
        responses.PATCH,
        "{}{}/".format(drivers_endpoint, driver_id),
        json=updated_postgresql_driver_server_resp,
    )

    driver = DataDriver.get(driver_id)
    driver.update(class_name="NewClassName", canonical_name="NewCanonicalName")
    assert driver.class_name == updated_postgresql_driver_server_resp["className"]
    assert driver.canonical_name == updated_postgresql_driver_server_resp["canonicalName"]


@responses.activate
def test_update_unable_to_change_base_names(
    drivers_endpoint, postgresql_driver_server_resp, mysql_driver_server_resp
):
    driver_id = postgresql_driver_server_resp["id"]
    responses.add(
        responses.GET,
        "{}{}/".format(drivers_endpoint, driver_id),
        json=postgresql_driver_server_resp,
    )

    responses.add(
        responses.PATCH,
        "{}{}/".format(drivers_endpoint, driver_id),
        json=postgresql_driver_server_resp,
    )

    driver = DataDriver.get(driver_id)

    driver._base_names = mysql_driver_server_resp["baseNames"]
    driver.update()

    # get driver from backend
    new_driver = DataDriver.get(driver_id)

    # assert server side has not been changed
    assert new_driver.base_names != mysql_driver_server_resp["baseNames"]
    assert new_driver.base_names == postgresql_driver_server_resp["baseNames"]


@responses.activate
def test_update_unable_to_change_creator(
    drivers_endpoint, postgresql_driver_server_resp, mysql_driver_server_resp
):
    driver_id = postgresql_driver_server_resp["id"]
    responses.add(
        responses.GET,
        "{}{}/".format(drivers_endpoint, driver_id),
        json=postgresql_driver_server_resp,
    )

    responses.add(
        responses.PATCH,
        "{}{}/".format(drivers_endpoint, driver_id),
        json=postgresql_driver_server_resp,
    )

    driver = DataDriver.get(driver_id)

    driver._creator = mysql_driver_server_resp["creator"]
    driver.update()

    # get driver from backend
    new_driver = DataDriver.get(driver_id)

    # assert server side has not been changed
    assert new_driver.creator != mysql_driver_server_resp["creator"]
    assert new_driver.creator == postgresql_driver_server_resp["creator"]


@responses.activate
def test_delete(drivers_endpoint, postgresql_driver_server_resp):
    driver_id = postgresql_driver_server_resp["id"]
    responses.add(
        responses.GET,
        "{}{}/".format(drivers_endpoint, driver_id),
        json=postgresql_driver_server_resp,
    )
    delete_url = "{}{}/".format(drivers_endpoint, driver_id)
    responses.add(responses.DELETE, delete_url)

    driver = DataDriver.get(driver_id)
    driver.delete()
    assert responses.calls[1].request.method == responses.DELETE
    assert responses.calls[1].request.url == delete_url



# encoding: utf-8
#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import json

from mock import Mock, patch
import pytest
import responses

from datarobot.errors import ClientError
from datarobot.models.project import Project
from datarobot.models.restore_discarded_features import DiscardedFeaturesInfo


@pytest.fixture
def async_url():
    return "https://host_name.com/status/status-id/"


@pytest.fixture
def discarded_features_url(unittest_endpoint, project_id):
    return "{}/projects/{}/discardedFeatures/".format(unittest_endpoint, project_id)


@pytest.fixture
def restore_discarded_features_url(unittest_endpoint, project_id):
    return "{}/projects/{}/modelingFeatures/fromDiscardedFeatures/".format(
        unittest_endpoint, project_id
    )


@pytest.fixture
def discarded_features_server_data():
    return {
        "totalRestoreLimit": 5,
        "remainingRestoreLimit": 3,
        "count": 3,
        "features": [
            "input4 (days from timestamp) (7 day mean)",
            "input4 (days from timestamp) (7 day min)",
            "input4 (days from timestamp) (7 day std)",
        ],
    }


@pytest.fixture
def discarded_features_response(discarded_features_url, discarded_features_server_data):
    responses.add(
        responses.GET,
        discarded_features_url,
        status=200,
        content_type="application/json",
        body=json.dumps(discarded_features_server_data),
    )


@pytest.fixture
def restore_discarded_features_response(restore_discarded_features_url, async_url):
    responses.add(
        responses.POST,
        restore_discarded_features_url,
        status=202,
        content_type="application/json",
        adding_headers={"Location": async_url},
        body=json.dumps(
            {
                "featuresToRestore": [
                    "input4 (days from timestamp) (7 day mean)",
                    "input4 (days from timestamp) (7 day min)",
                ],
                "warnings": ['"ololo" does not exist'],
            }
        ),
    )


@pytest.fixture
def discarded_features_not_found_response(discarded_features_url):
    responses.add(
        responses.GET,
        discarded_features_url,
        status=404,
        content_type="application/json",
        body=json.dumps({"message": "no discarded features info"}),
    )


@pytest.fixture
def discarded_features_not_allowed_response(discarded_features_url):
    responses.add(
        responses.GET,
        discarded_features_url,
        status=422,
        content_type="application/json",
        body=json.dumps({"message": "Segmented project is not allowed"}),
    )


@responses.activate
@pytest.mark.usefixtures("discarded_features_response")
def test_retrieve_discarded_features(project_id):
    project = Project(project_id)
    discarded_feature_info = project.get_discarded_features()
    assert discarded_feature_info.features == [
        "input4 (days from timestamp) (7 day mean)",
        "input4 (days from timestamp) (7 day min)",
        "input4 (days from timestamp) (7 day std)",
    ]
    assert discarded_feature_info.count == len(discarded_feature_info.features)
    assert discarded_feature_info.remaining_restore_limit == 3
    assert discarded_feature_info.total_restore_limit == 5


@responses.activate
@pytest.mark.usefixtures("discarded_features_not_found_response")
def test_retrieve_no_discarded_features(project_id):
    with pytest.raises(ClientError):
        DiscardedFeaturesInfo.retrieve(project_id)


@responses.activate
@pytest.mark.usefixtures("discarded_features_not_allowed_response")
def test_retrieve_segmented_project_not_allowed(project_id):
    with pytest.raises(ClientError):
        DiscardedFeaturesInfo.retrieve(project_id)


@responses.activate
@pytest.mark.usefixtures("restore_discarded_features_response")
@patch("datarobot.models.restore_discarded_features.wait_for_async_resolution", Mock())
def test_restore_discarded_features(project_id):
    project = Project(project_id)
    restored_info = project.restore_discarded_features(
        ["input4 (days from timestamp) (7 day mean)", "input4 (days from timestamp) (7 day min)"]
    )
    assert restored_info.warnings == ['"ololo" does not exist']
    assert restored_info.restored_features == [
        "input4 (days from timestamp) (7 day mean)",
        "input4 (days from timestamp) (7 day min)",
    ]



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import copy
import json

import responses

from datarobot._experimental.models.model_package import ModelPackage
from tests.utils import add_response

_base_url = "https://host_name.com"
_project_id = "60feb853eadeed4617dd5de8"
_model_id = "60feb9286c9dfb8b69b93c80"
_model_package_id = "60feb9b5eadeed4617dd5e29"

_create_url = "{}/modelPackages/fromLearningModel/".format(_base_url)
_build_url = "{}/modelPackages/{}/modelPackageFileBuilds/".format(_base_url, _model_package_id)
_download_url = "{}/modelPackages/{}/modelPackageFile/".format(_base_url, _model_package_id)
_get_url = "{}/modelPackages/{}/".format(_base_url, _model_package_id)
_list_url = "{}/modelPackages/".format(_base_url)

_model_package_json = {
    "id": _model_package_id,
    "name": "model package name",
    "modelId": _model_id,
    "modelExecutionType": "dedicated",
    "activeDeploymentCount": 0,
    "isArchived": False,
    "permissions": ["CAN_VIEW", "CAN_EDIT", "CAN_SHARE", "CAN_ARCHIVE", "CAN_SHARE_OWNERSHIP"],
    "importMeta": {
        "creatorId": "60f82edf3cf8c8184476ee46",
        "creatorUsername": "admin@datarobot.com",
        "dateCreated": "2021-07-26T13:33:41.483782Z",
        "originalFileName": None,
    },
    "modelKind": {
        "isTimeSeries": True,
        "isMultiseries": True,
        "isUnsupervisedLearning": True,
        "isAnomalyDetectionModel": False,
        "isFeatureDiscovery": False,
    },
    "target": {
        "name": "DR_FAKE_TARGET (actual)",
        "type": "Multiclass",
        "classNames": ["Cluster 2", "Cluster 1"],
        "classCount": 2,
        "predictionThreshold": None,
        "predictionProbabilitiesColumn": None,
    },
    "modelDescription": {
        "modelName": "model name",
        "description": "",
        "location": "",
        "buildEnvironmentType": "DataRobot",
    },
    "datasets": {
        "datasetName": "dataset.csv",
        "trainingDataCatalogId": None,
        "baselineSegmentedBy": None,
        "holdoutDataCatalogId": None,
        "holdoutDatasetName": None,
    },
    "sourceMeta": {
        "projectId": _project_id,
        "projectName": "project name",
        "environmentUrl": "https://host_name.com",
        "scoringCode": {"location": None, "dataRobotPredictionVersion": "2.1.11"},
    },
    "timeseries": {
        "datetimeColumnName": "Date (actual)",
        "datetimeColumnFormat": "%Y-%m-%d",
        "forecastDistances": [0],
        "forecastDistancesTimeUnit": "DAY",
        "forecastDistanceColumnName": "Forecast Distance",
        "forecastPointColumnName": "dr_forecast_point",
        "seriesColumnName": "Series",
        "featureDerivationWindowStart": -35,
        "featureDerivationWindowEnd": 0,
        "effectiveFeatureDerivationWindowStart": -35,
        "effectiveFeatureDerivationWindowEnd": 0,
        "isTraditionalTimeSeries": False,
        "isCrossSeries": False,
        "isNewSeriesSupport": False,
    },
}


@responses.activate
def test_model_package_create():
    responses.add(
        responses.POST, _create_url, body=json.dumps(_model_package_json), status=200,
    )

    model_package = ModelPackage.create(_model_package_id)
    assert model_package.id == _model_package_id


@responses.activate
def test_model_package_get():
    add_response(url=_get_url, body=_model_package_json)

    model_package = ModelPackage.get(_model_package_id)
    assert model_package.id == _model_package_id


@responses.activate
def test_model_package_list():
    dummy_model_package_json = copy.deepcopy(_model_package_json)
    dummy_model_package_json["id"] = "60feb9b5eadeed4617dd5e33"

    add_response(
        url=_list_url,
        body={
            "count": 1,
            "next": None,
            "previous": None,
            "data": [_model_package_json, dummy_model_package_json],
        },
    )

    model_package_list = ModelPackage.list()
    assert len(model_package_list) == 2
    first_model_package = model_package_list[0]
    assert first_model_package.id == _model_package_id
    second_model_package = model_package_list[1]
    assert second_model_package.id == "60feb9b5eadeed4617dd5e33"



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import os
import tempfile
import unittest

import mock
import pytest
import responses
import six

import datarobot as dr
from datarobot.client import Client, get_client, set_client
from datarobot.enums import DEFAULT_TIMEOUT
from datarobot.errors import AppPlatformError, ClientError
from datarobot.rest import RESTClientObject, Retry
from tests.utils import SDKTestcase

if six.PY3:
    from unittest.mock import patch
else:
    from mock import patch


@pytest.mark.no_client
class TestClient(object):
    @pytest.mark.usefixtures("no_client_version_check")
    def test_instantiation(self):
        """
        Basic client installation.
        """
        client = Client(token="t-token", endpoint="https://host_name.com")
        assert get_client() is client

    @pytest.mark.usefixtures("no_client_version_check")
    @mock.patch("datarobot.client._file_exists", return_value=False)
    def test_instantiation_without_env(self, _):
        """
        Basic client installation by get_client without configuration.
        """
        with mock.patch("os.environ", {"DATAROBOT_ENDPOINT": "https://host_name.com"}):
            with pytest.raises(ValueError):
                get_client()

    @pytest.mark.usefixtures("no_client_version_check")
    @mock.patch("datarobot.client._file_exists")
    def test_no_endpoint_fails(self, mock_file_exists):
        mock_file_exists.return_value = False
        with mock.patch("os.environ", {}):
            with pytest.raises(ValueError):
                Client(token="NOTATOKEN")

    @pytest.mark.usefixtures("no_client_version_check")
    def test_token_and_endpoint_is_okay(self, no_client_version_check):
        Client(token="token", endpoint="https://need_an_endpoint.com")
        assert no_client_version_check.called

    def test_bad_endpoint_fails(self):
        """
        Client with bad endpoint
        """
        with mock.patch("datarobot.rest.RESTClientObject.get") as mock_version_check:
            mock_version_check.side_effect = ClientError("dummy msg", 12345)
            with pytest.raises(UserWarning, match="Server did not reply"):
                Client(token="t-token", endpoint="https://host_name.com")

    @pytest.mark.usefixtures("no_client_version_check")
    def test_re_instantiation(self):
        """
        Client re installation.
        """
        previous = Client(token="t-****", endpoint="https://end.com")
        old_client = set_client(
            RESTClientObject(auth=("u-**********", "p-******"), endpoint="https://host_name.com")
        )
        assert previous is old_client

    @pytest.mark.usefixtures("no_client_version_check")
    @responses.activate
    def test_recognizing_domain_on_instance(self):
        raw = """{"api_token": "some_token"}"""
        responses.add(
            responses.POST,
            "https://host_name.com/api/v2/api_token/",
            body=raw,
            status=201,
            content_type="application/json",
        )
        set_client(
            RESTClientObject(
                auth=("u-**********", "p-******"), endpoint="https://host_name.com/api/v2"
            )
        )
        restored_client = get_client()
        assert restored_client.domain == "https://host_name.com"

    @pytest.mark.usefixtures("no_client_version_check")
    @mock.patch("datarobot.client._file_exists", return_value=False)
    def test_instantiation_from_env(self, mock_file_exists):
        """
        Test instantiation with creds from virtual environment
        """
        with patch.dict(
            "os.environ",
            {"DATAROBOT_API_TOKEN": "venv_token", "DATAROBOT_ENDPOINT": "https://host_name.com"},
        ):
            rest_client = get_client()
            assert rest_client.token == "venv_token"
            assert rest_client.endpoint == "https://host_name.com"

        set_client(None)

        with patch.dict(
            "os.environ",
            {
                "DATAROBOT_API_TOKEN": "venv_token",
                "DATAROBOT_ENDPOINT": "https://host_name.com",
                "DATAROBOT_MAX_RETRIES": "2",
            },
        ):
            rest_client = get_client()
            assert rest_client.token == "venv_token"
            assert rest_client.connect_timeout == DEFAULT_TIMEOUT.CONNECT
            assert rest_client.adapters
            for adapter in rest_client.adapters.values():
                assert adapter.max_retries.total == 2

    @pytest.mark.usefixtures("no_client_version_check")
    @mock.patch("datarobot.client._file_exists", return_value=False)
    def test_instantiation_from_env_checks_client_compatibility(
        self, mock_file_exists, no_client_version_check
    ):
        """
        Test instantiation with creds from virtual environment checks client compatibility.
        """
        with patch.dict(
            "os.environ",
            {"DATAROBOT_API_TOKEN": "venv_token", "DATAROBOT_ENDPOINT": "https://host_name.com"},
        ):
            rest_client = Client()
            assert rest_client.token == "venv_token"
            assert rest_client.endpoint == "https://host_name.com"
        assert no_client_version_check.called

    @pytest.mark.usefixtures("no_client_version_check")
    def test_instantiation_from_file_with_wrong_path(self):
        with patch.dict("os.environ", {"DATAROBOT_CONFIG_FILE": "./tests/fixtures/.datarobotrc"}):
            with pytest.raises(ValueError):
                get_client()

    @pytest.mark.usefixtures("no_client_version_check")
    def test_instantiation_from_yaml_file_api_token(self):
        file_data = "token: fake_token\nendpoint: https://host_name.com\nmax_retries: 3"
        config_file = tempfile.NamedTemporaryFile(delete=False, suffix=".yaml")
        try:
            with open(config_file.name, mode="w") as config:
                config.write(file_data)
            fake_environ = {"DATAROBOT_CONFIG_FILE": config_file.name}
            with patch("os.environ", fake_environ):
                rest_client = get_client()
            assert rest_client.token == "fake_token"
            assert rest_client.endpoint == "https://host_name.com"
            assert rest_client.connect_timeout == DEFAULT_TIMEOUT.CONNECT
            assert rest_client.adapters
            for adapter in rest_client.adapters.values():
                assert adapter.max_retries.total == 3
        finally:
            os.remove(config_file.name)

    @pytest.mark.usefixtures("no_client_version_check")
    def test_instantiation_from_yaml_file_set_timeout(self, temporary_file):
        connect_timeout = 55
        file_data = (
            "token: fake_token\nendpoint: https://host_name.com\nconnect_timeout: {}"
        ).format(connect_timeout)
        with open(temporary_file, mode="w") as config:
            config.write(file_data)
        fake_environ = {"DATAROBOT_CONFIG_FILE": temporary_file}
        with patch("os.environ", fake_environ):
            rest_client = get_client()
        assert rest_client.connect_timeout == connect_timeout

    @pytest.mark.usefixtures("no_client_version_check")
    def test_instantiation_from_yaml_checks_client_compatibility(
        self, temporary_file, no_client_version_check
    ):
        connect_timeout = 55
        file_data = (
            "token: fake_token\nendpoint: https://host_name.com\nconnect_timeout: {}"
        ).format(connect_timeout)
        with open(temporary_file, mode="w") as config:
            config.write(file_data)
        fake_environ = {"DATAROBOT_CONFIG_FILE": temporary_file}
        with patch("os.environ", fake_environ):
            Client()
        assert no_client_version_check.called

    @pytest.mark.usefixtures("no_client_version_check")
    def test_client_from_codeline(self):
        Client(token="some_token", endpoint="https://endpoint.com")
        c = get_client()
        assert c.token == "some_token"
        assert c.endpoint == "https://endpoint.com"
        assert c.connect_timeout == DEFAULT_TIMEOUT.CONNECT
        assert c.adapters
        for adapter in c.adapters.values():
            assert adapter.max_retries.total > 0
            assert adapter.max_retries.method_whitelist == Retry.DEFAULT_METHOD_WHITELIST
            assert adapter.max_retries.respect_retry_after_header

    @pytest.mark.usefixtures("no_client_version_check")
    def test_client_from_codeline_set_timeout(self):
        connect_timeout = 55
        Client(token="some_token", endpoint="https://endpoint.com", connect_timeout=connect_timeout)
        c = get_client()
        assert c.connect_timeout == connect_timeout

    @pytest.mark.usefixtures("no_client_version_check")
    def test_client_from_codeline_max_retries(self):
        Client(
            token="some token",
            endpoint="https://endpoint.com",
            max_retries=Retry(read=3, connect=3),
        )
        c = get_client()
        assert c.adapters
        for adapter in c.adapters.values():
            assert adapter.max_retries.read == 3
            assert adapter.max_retries.connect == 3

    @pytest.mark.usefixtures("no_client_version_check")
    def test_client_user_agent(self):
        with patch("platform.system") as mock_system, patch(
            "platform.release"
        ) as mock_release, patch("platform.machine") as mock_machine, patch(
            "platform.python_version"
        ) as mock_py_version:
            mock_system.return_value = "Prince"
            mock_release.return_value = "Party"
            mock_machine.return_value = "x19_99"
            mock_py_version.return_value = "3.14.159"
            header = RESTClientObject._make_user_agent_header()
        expected_agent = "DataRobotPythonClient/{} (Prince Party x19_99) Python-3.14.159".format(
            dr.__version__
        )
        assert header == {"User-Agent": expected_agent}

    @pytest.mark.usefixtures("no_client_version_check")
    def test_client_user_agent_with_suffix(self):
        suffix = "Blamo/8.23 (6.30)"
        orig = RESTClientObject._make_user_agent_header()
        with_suffix = RESTClientObject._make_user_agent_header(suffix=suffix)
        assert with_suffix["User-Agent"].startswith(orig["User-Agent"])
        assert with_suffix["User-Agent"].endswith(" " + suffix)


class RestErrors(SDKTestcase):
    @responses.activate
    def test_404_plain_text(self):
        """
        Bad request in plain text
        """
        raw = "Not Found"

        responses.add(
            responses.GET,
            "https://host_name.com/projects/404",
            body=raw,
            status=404,
            content_type="text/plain",
        )

        client = get_client()

        with self.assertRaises(AppPlatformError) as app_error:
            client.get("projects/404")

        self.assertEqual(str(app_error.exception), "404 client error: Not Found")

    @responses.activate
    def test_404_json(self):
        """
        Bad request with datarobot reason in json
        """
        raw = """
        {"message": "Not Found"}
        """

        responses.add(
            responses.GET,
            "https://host_name.com/projects/404",
            body=raw,
            status=404,
            content_type="application/json",
        )

        client = get_client()

        with self.assertRaises(AppPlatformError) as app_error:
            client.get("projects/404")

        exception_message = str(app_error.exception)
        self.assertIn("404 client error", exception_message)
        self.assertIn("Not Found", exception_message)

    @responses.activate
    def test_500_json(self):
        """
        Bad request with datarobot reason in json
        """
        raw = """
        {"message": "Not Found"}
        """

        responses.add(
            responses.GET,
            "https://host_name.com/projects/500",
            body=raw,
            status=500,
            content_type="application/json",
        )

        client = get_client()

        with self.assertRaises(AppPlatformError) as app_error:
            client.get("projects/500")

        exception_message = str(app_error.exception)
        self.assertIn("500 server error", exception_message)
        self.assertIn("Not Found", exception_message)

    @responses.activate
    def test_other_errors(self):
        """
        Other errors
        """
        raw = """
        {"message": "Bad response"}
        """

        responses.add(
            responses.GET,
            "https://host_name.com/projects/500",
            body=raw,
            status=500,
            content_type="application/json",
        )

        client = get_client()

        with self.assertRaises(AppPlatformError) as app_error:
            client.get("projects/500")
            self.assertEqual(
                str(app_error.exception), "Connection refused: https://host_name.com/projects/500",
            )


class TestClientAttributes(unittest.TestCase):
    def test_main_useful_things_under_datarobot(self):
        """To lower the intimidation factor, let's try to limit the objects
        that show up at the root of datarobot

        This way, when they are in IPython and do tab-completion they get a
        sense for what is available to tinker with at the top level
        """
        known_names = {
            "Project",
            "Model",
            "DatetimeModel",
            "PrimeModel",
            "BlenderModel",
            "FrozenModel",
            "Ruleset",
            "Blueprint",
            "BlueprintTaskDocument",
            "BlueprintChart",
            "ModelBlueprintChart",
            "ModelJob",
            "PredictJob",
            "Job",
            "Predictions",
            "PredictionDataset",
            "PrimeFile",
            "QUEUE_STATUS",
            "Client",
            "AUTOPILOT_MODE",
            "AppPlatformError",
            "utils",
            "errors",
            "models",
            "client",
            "rest",
            "SCORING_TYPE",
            "DatasetFeaturelist",
            "Featurelist",
            "ModelingFeaturelist",
            "Feature",
            "FeatureHistogram",
            "FeatureSettings",
            "ModelingFeature",
            "FeatureLineage",
            "helpers",
            "RandomCV",
            "StratifiedCV",
            "GroupCV",
            "UserCV",
            "RandomTVH",
            "UserTVH",
            "StratifiedTVH",
            "GroupTVH",
            "DatetimePartitioning",
            "DatetimePartitioningSpecification",
            "BacktestSpecification",
            "partitioning_methods",
            "AdvancedOptions",
            "Periodicity",
            "VERBOSITY_LEVEL",
            "enums",
            "ImportedModel",
            "ReasonCodesInitialization",
            "ReasonCodes",
            "PayoffMatrix",
            "PredictionExplanationsInitialization",
            "PredictionExplanations",
            "RatingTable",
            "RatingTableModel",
            "TARGET_TYPE",
            "TrainingPredictions",
            "TrainingPredictionsJob",
            "ShapImpact",
            "ShapMatrix",
            "ShapMatrixJob",
            "SharingAccess",
            "ModelRecommendation",
            "DataDriver",
            "Connector",
            "DataStore",
            "Dataset",
            "DatasetDetails",
            "DatasetFeature",
            "DatasetFeatureHistogram",
            "DataSource",
            "DataSourceParameters",
            "ComplianceDocumentation",
            "ComplianceDocTemplate",
            "CalendarFile",
            "PredictionServer",
            "Deployment",
            "SecondaryDatasetConfigurations",
            "BatchPredictionJob",
            "BatchPredictionJobDefinition",
            "Credential",
            "InteractionFeature",
            "ExternalScores",
            "ExternalLiftChart",
            "ExternalResidualsChart",
            "ExternalRocCurve",
            "ExternalMulticlassLiftChart",
            "ExternalConfusionChart",
            "ExecutionEnvironment",
            "ExecutionEnvironmentVersion",
            "CustomInferenceImage",
            "CustomInferenceModel",
            "CustomModelVersion",
            "CustomModelVersionDependencyBuild",
            "CustomModelTest",
            "CustomTask",
            "CustomTaskVersion",
            "FeatureImpactJob",
            "RelationshipsConfiguration",
            "NETWORK_EGRESS_POLICY",
            "FeatureAssociationMatrix",
            "FeatureAssociationMatrixDetails",
            "FeatureAssociationFeaturelists",
            "AutomatedDocument",
            "DatasetDefinition",
            "feature_discovery",
            "SecondaryDataset",
            "Relationship",
            "FeatureDiscoverySetting",
            "SNAPSHOT_POLICY",
            "UserBlueprint",
            "DataEngineQueryGenerator",
            "ClassMappingAggregationSettings",
            "ClusteringModel",
            "ClusterInsight",
            "Cluster",
        }

        found_names = {name for name in dir(dr) if not name.startswith("_")}
        assert found_names == known_names


@pytest.yield_fixture
def mock_os():
    with patch("datarobot.client.os", autospec=True) as m:
        yield m


@pytest.yield_fixture
def mock_file_exists():
    with patch("datarobot.client._file_exists") as m:
        yield m


@pytest.yield_fixture
def mock_os_environ():
    with patch("datarobot.client.os.environ") as m:
        yield m



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from datarobot.models.model import ClusteringModel
from datarobot.utils import from_api


class TestClusteringModel(object):
    def test_class_creation(self, clustering_model_data):
        model = ClusteringModel.from_data(from_api(clustering_model_data))
        # assert new fields specific to clustering models
        assert clustering_model_data["nClusters"] == model.n_clusters
        assert (
            clustering_model_data["isNClustersDynamicallyDetermined"]
            == model.is_n_clusters_dynamically_determined
        )



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import json
import os

import contextlib2
from requests_toolbelt import MultipartEncoder
import trafaret as t

from datarobot._compat import Int, String
from datarobot.models.api_object import APIObject
from datarobot.models.custom_model_version import (
    CustomDependency,
    CustomModelFileItem,
    RequiredMetadataValue,
)
from datarobot.models.custom_task_version_dependency_build import CustomTaskVersionDependencyBuild
from datarobot.models.trafarets import UserBlueprintTaskArgument_
from datarobot.models.user_blueprints.models import UserBlueprintTaskArgument
from datarobot.utils import camelize, deprecation_warning, encode_utf8_if_py2
from datarobot.utils.pagination import unpaginate
from datarobot.utils.waiters import wait_for_custom_resolution


class CustomTaskFileItem(CustomModelFileItem):
    """A file item attached to a DataRobot custom task version.

    .. versionadded:: v2.26

    Attributes
    ----------
    id: str
        id of the file item
    file_name: str
        name of the file item
    file_path: str
        path of the file item
    file_source: str
        source of the file item
    created_at: str
        ISO-8601 formatted timestamp of when the version was created
    """

    _converter = t.Dict(
        {
            t.Key("id"): String(),
            t.Key("file_name"): String(),
            t.Key("file_path"): String(),
            t.Key("file_source"): String(),
            t.Key("created") >> "created_at": String(),
        }
    ).ignore_extra("*")

    schema = _converter


class CustomTaskVersion(APIObject):
    """A version of a DataRobot custom task.

    .. versionadded:: v2.26

    Attributes
    ----------
    id: str
        id of the custom task version
    custom_task_id: str
        id of the custom task
    version_minor: int
        a minor version number of custom task version
    version_major: int
        a major version number of custom task version
    label: str
        short human readable string to label the version
    created_at: str
        ISO-8601 formatted timestamp of when the version was created
    is_frozen: bool
        a flag if the custom task version is frozen
    items: List[CustomTaskFileItem]
        a list of file items attached to the custom task version
    description: str, optional
        custom task version description
    required_metadata: Dict[str, str]
        Deprecated in v2.26. Use require_metadata_values instead.

        Additional parameters required by the execution environment. The required keys are
        defined by the fieldNames in the base environment's requiredMetadataKeys. Once set,
        they cannot be changed. If you want to change them, make a new version.
    base_environment_id: str, optional
        id of the environment to use with the task
    base_environment_version_id: str, optional
        id of the environment version to use with the task
    dependencies: List[CustomDependency]
        the parsed dependencies of the custom task version if the
        version has a valid requirements.txt file
    required_metadata_values: List[RequiredMetadataValue]
        Additional parameters required by the execution environment. The required keys are
        defined by the fieldNames in the base environment's requiredMetadataKeys.
    arguments: List[UserBlueprintTaskArgument]
        A list of custom task version arguments.
    """

    _path = "customTasks/{}/versions/"
    _dependency_build_path = "customTasks/{}/versions/{}/dependencyBuild/"
    _dependency_build_log_path = "customTasks/{}/versions/{}/dependencyBuildLog/"

    _converter = t.Dict(
        {
            t.Key("id"): String(),
            t.Key("custom_task_id"): String(),
            t.Key("version_major"): Int(),
            t.Key("version_minor"): Int(),
            t.Key("label"): String(),
            t.Key("created") >> "created_at": String(),
            t.Key("is_frozen"): t.Bool(),
            t.Key("items"): t.List(CustomTaskFileItem.schema),
            # because `from_server_data` scrubs Nones, this must be optional here.
            t.Key("description", optional=True): String(max_length=10000, allow_blank=True)
            | t.Null(),
            t.Key("required_metadata", optional=True): t.Mapping(String(), String()),
            t.Key("base_environment_id", optional=True): String() | t.Null(),
            t.Key("base_environment_version_id", optional=True): String() | t.Null(),
            t.Key("dependencies", optional=True): t.List(CustomDependency.schema),
            t.Key("required_metadata_values", optional=True): t.List(RequiredMetadataValue.schema),
            t.Key("arguments", optional=True): t.List(UserBlueprintTaskArgument_),
        }
    ).ignore_extra("*")

    schema = _converter

    def __init__(
        self,
        id,
        custom_task_id,
        version_major,
        version_minor,
        label,
        created_at,
        is_frozen,
        items,
        description=None,
        required_metadata=None,
        base_environment_id=None,
        base_environment_version_id=None,
        dependencies=None,
        required_metadata_values=None,
        arguments=None,
    ):
        if dependencies is None:
            dependencies = []
        arguments = arguments or []

        self.id = id
        self.custom_task_id = custom_task_id
        self.description = description
        self.version_major = version_major
        self.version_minor = version_minor
        self.label = label
        self.created_at = created_at
        self.is_frozen = is_frozen
        self.items = [CustomTaskFileItem(**data) for data in items]

        self.required_metadata = required_metadata
        self.required_metadata_values = (
            [RequiredMetadataValue(**val) for val in required_metadata_values]
            if required_metadata_values
            else None
        )
        self.base_environment_id = base_environment_id
        self.base_environment_version_id = base_environment_version_id
        self.dependencies = [CustomDependency(**data) for data in dependencies]
        self.arguments = [UserBlueprintTaskArgument(**argument) for argument in arguments]

    def __repr__(self):
        return encode_utf8_if_py2(
            u"{}({!r})".format(self.__class__.__name__, self.label or self.id)
        )

    def _update_values(self, new_response):
        # type (CustomTaskVersion) -> None
        for attr in self._fields():
            new_value = getattr(new_response, attr)
            setattr(self, attr, new_value)

    @classmethod
    def _all_versions_path(cls, task_id):
        # type: (str) -> str
        return cls._path.format(task_id)

    @classmethod
    def _single_version_path(cls, task_id, version_id):
        # type: (str, str) -> str
        return cls._all_versions_path(task_id) + "{}/".format(version_id)

    @classmethod
    def from_server_data(cls, data, keep_attrs=None):
        initial = super(CustomTaskVersion, cls).from_server_data(data, keep_attrs)
        # from_server_data will make the keys in requiredMetadata lowercase,
        # which is not OK. we need to preserve case
        initial.required_metadata = data.get("requiredMetadata")
        return initial

    @classmethod
    def create_clean(
        cls,
        custom_task_id,
        base_environment_id,
        is_major_update=True,
        folder_path=None,
        required_metadata=None,
        required_metadata_values=None,
    ):
        """Create a custom task version without files from previous versions.

        .. versionadded:: v2.26

        Parameters
        ----------
        custom_task_id: str
            the id of the custom task
        base_environment_id: str
            the id of the base environment to use with the custom task version
        is_major_update: bool, optional
            if the current version is 2.3, `True` would set the new version at `3.0`.
            `False` would set the new version at `2.4`.
            Default to `True`
        folder_path: str, optional
            the path to a folder containing files to be uploaded.
            Each file in the folder is uploaded under path relative
            to a folder path
        required_metadata: Dict[str, str]
            Deprecated in v2.26. Use require_metadata_values instead.

            Additional parameters required by the execution environment. The required keys are
            defined by the fieldNames in the base environment's requiredMetadataKeys. Once set,
            they cannot be changed. If you want to change them, make a new version.
        required_metadata_values: List[RequiredMetadataValue]
            Additional parameters required by the execution environment. The required keys are
            defined by the fieldNames in the base environment's requiredMetadataKeys.

        Returns
        -------
        CustomTaskVersion
            created custom task version

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        return cls._create(
            "post",
            custom_task_id,
            is_major_update,
            base_environment_id,
            folder_path,
            required_metadata=required_metadata,
            required_metadata_values=required_metadata_values,
        )

    @classmethod
    def create_from_previous(
        cls,
        custom_task_id,
        base_environment_id,
        is_major_update=True,
        folder_path=None,
        files_to_delete=None,
        required_metadata=None,
        required_metadata_values=None,
    ):
        """Create a custom task version containing files from a previous version.

        .. versionadded:: v2.26

        Parameters
        ----------
        custom_task_id: str
            the id of the custom task
        base_environment_id: str
            the id of the base environment to use with the custom task version
        is_major_update: bool, optional
            if the current version is 2.3, `True` would set the new version at `3.0`.
            `False` would set the new version at `2.4`.
            Default to `True`
        folder_path: str, optional
            the path to a folder containing files to be uploaded.
            Each file in the folder is uploaded under path relative
            to a folder path
        files_to_delete: list, optional
            the list of a file items ids to be deleted
            Example: ["5ea95f7a4024030aba48e4f9", "5ea6b5da402403181895cc51"]
        required_metadata: Dict[str, str]
            Deprecated in v2.26. Use require_metadata_values instead.

            Additional parameters required by the execution environment. The required keys are
            defined by the fieldNames in the base environment's requiredMetadataKeys. Once set,
            they cannot be changed. If you want to change them, make a new version.
        required_metadata_values: List[RequiredMetadataValue]
            Additional parameters required by the execution environment. The required keys are
            defined by the fieldNames in the base environment's requiredMetadataKeys.

        Returns
        -------
        CustomTaskVersion
            created custom task version

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        return cls._create(
            "patch",
            custom_task_id,
            is_major_update,
            base_environment_id,
            folder_path,
            files_to_delete=files_to_delete,
            required_metadata=required_metadata,
            required_metadata_values=required_metadata_values,
        )

    @classmethod
    def _create(
        cls,
        method,
        custom_task_id,
        is_major_update,
        base_environment_id,
        folder_path=None,
        files_to_delete=None,
        required_metadata=None,
        required_metadata_values=None,
    ):
        url = cls._all_versions_path(custom_task_id)

        upload_data = [
            ("isMajorUpdate", str(is_major_update)),
            ("baseEnvironmentId", base_environment_id),
        ]
        if files_to_delete:
            upload_data += [("filesToDelete", file_id) for file_id in files_to_delete]
        if required_metadata is not None:
            deprecation_warning(
                "required_metadata",
                deprecated_since_version="2.26",
                will_remove_version="2.27",
                message="Use required_metadata_values instead.",
            )
            if required_metadata_values is not None:
                raise ValueError(
                    "Cannot specify both required_metadata and required_metadata_values. "
                    "Please use required_metadata_values."
                )
            upload_data.append(("requiredMetadata", json.dumps(required_metadata)))

        if required_metadata_values is not None:
            upload_data.append(
                (
                    "requiredMetadataValues",
                    json.dumps(
                        [
                            {camelize(k): v for k, v in val.to_dict().items()}
                            for val in required_metadata_values
                        ]
                    ),
                )
            )

        cls._verify_folder_path(folder_path)

        with contextlib2.ExitStack() as stack:
            if folder_path:
                for dir_name, _, file_names in os.walk(folder_path):
                    for file_name in file_names:
                        file_path = os.path.join(dir_name, file_name)
                        file = stack.enter_context(open(file_path, "rb"))

                        upload_data.append(("file", (os.path.basename(file_path), file)))
                        upload_data.append(("filePath", os.path.relpath(file_path, folder_path)))

            encoder = MultipartEncoder(fields=upload_data)
            headers = {"Content-Type": encoder.content_type}
            response = cls._client.request(method, url, data=encoder, headers=headers)
        return cls.from_server_data(response.json())

    @staticmethod
    def _verify_folder_path(folder_path):
        if folder_path and not os.path.exists(folder_path):
            raise ValueError("The folder: {} does not exist.".format(folder_path))

    @classmethod
    def list(cls, custom_task_id):
        """List custom task versions.

        .. versionadded:: v2.26

        Parameters
        ----------
        custom_task_id: str
            the id of the custom task

        Returns
        -------
        List[CustomTaskVersion]
            a list of custom task versions

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        url = cls._all_versions_path(custom_task_id)
        data = unpaginate(url, None, cls._client)
        return [cls.from_server_data(item) for item in data]

    @classmethod
    def get(cls, custom_task_id, custom_task_version_id):
        """Get custom task version by id.

        .. versionadded:: v2.26

        Parameters
        ----------
        custom_task_id: str
            the id of the custom task
        custom_task_version_id: str
            the id of the custom task version to retrieve

        Returns
        -------
        CustomTaskVersion
            retrieved custom task version

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status.
        datarobot.errors.ServerError
            if the server responded with 5xx status.
        """
        path = cls._single_version_path(custom_task_id, custom_task_version_id)
        return cls.from_location(path)

    def download(self, file_path):
        """Download custom task version.

        .. versionadded:: v2.26

        Parameters
        ----------
        file_path: str
            path to create a file with custom task version content

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status.
        datarobot.errors.ServerError
            if the server responded with 5xx status.
        """

        response = self._client.get(
            self._single_version_path(self.custom_task_id, self.id) + "download/"
        )
        with open(file_path, "wb") as f:
            f.write(response.content)

    def update(self, description=None, required_metadata=None, required_metadata_values=None):
        """Update custom task version properties.

        .. versionadded:: v2.26

        Parameters
        ----------
        description: str
            new custom task version description
        required_metadata: Dict[str, str]
            Deprecated in v2.26. Use require_metadata_values instead.

            Additional parameters required by the execution environment. The required keys are
            defined by the fieldNames in the base environment's requiredMetadataKeys. Once set,
            they cannot be changed. If you want to change them, make a new version.
        required_metadata_values: List[RequiredMetadataValue]
            Additional parameters required by the execution environment. The required keys are
            defined by the fieldNames in the base environment's requiredMetadataKeys.

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status.
        datarobot.errors.ServerError
            if the server responded with 5xx status.
        """
        payload = {}
        if description:
            payload.update({"description": description})
        if required_metadata is not None:
            deprecation_warning(
                "required_metadata",
                deprecated_since_version="2.26",
                will_remove_version="2.27",
                message="Use required_metadata_values instead.",
            )
            if required_metadata_values is not None:
                raise ValueError(
                    "Cannot specify both required_metadata and required_metadata_values. "
                    "Please use required_metadata_values."
                )
            payload.update({"requiredMetadata": required_metadata})

        if required_metadata_values is not None:
            payload.update(
                {"requiredMetadataValues": [val.to_dict() for val in required_metadata_values]}
            )

        url = self._path.format(self.custom_task_id)
        path = "{}{}/".format(url, self.id)

        response = self._client.patch(path, data=payload)

        data = response.json()
        new_version = CustomTaskVersion.from_server_data(data)
        self._update_values(new_version)

    def refresh(self):
        """Update custom task version with the latest data from server.

        .. versionadded:: v2.26

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """

        new_object = self.get(self.custom_task_id, self.id)
        self._update_values(new_object)

    def start_dependency_build(self):
        """Start the dependency build for a custom task version and return build status.
        .. versionadded:: v2.27

        Returns
        -------
        CustomTaskVersionDependencyBuild
            DTO of custom task version dependency build.
        """
        url = self._dependency_build_path.format(self.custom_task_id, self.id)
        response = self._client.post(url)
        server_data = response.json()
        return CustomTaskVersionDependencyBuild.from_server_data(server_data)

    def start_dependency_build_and_wait(self, max_wait):
        """Start the dependency build for a custom task version and wait while pulling status.
        .. versionadded:: v2.27

        Parameters
        ----------
        max_wait: int
            max time to wait for a build completion

        Returns
        -------
        CustomTaskVersionDependencyBuild
            DTO of custom task version dependency build.

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        datarobot.errors.AsyncTimeoutError
            Raised if the dependency build is not finished after max_wait.
        """
        custom_task_id = self.custom_task_id
        custom_task_version_id = self.id

        def build_complete(response):
            server_data = response.json()
            if server_data["buildStatus"] in ["success", "failed"]:
                return CustomTaskVersionDependencyBuild.from_server_data(server_data)
            return None

        url = self._dependency_build_path.format(custom_task_id, custom_task_version_id)
        self._client.post(url)
        return wait_for_custom_resolution(self._client, url, build_complete, max_wait)

    def cancel_dependency_build(self):
        """Cancel custom task version dependency build that is in progress.
        .. versionadded:: v2.27

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """

        url = self._dependency_build_path.format(self.custom_task_id, self.id)
        self._client.delete(url)

    def get_dependency_build(self):
        """Retrieve information about a custom task version's dependency build.
        .. versionadded:: v2.27

        Returns
        -------
        CustomTaskVersionDependencyBuild
            DTO of custom task version dependency build.
        """

        url = self._dependency_build_path.format(self.custom_task_id, self.id)
        response = self._client.get(url)
        server_data = response.json()
        return CustomTaskVersionDependencyBuild.from_server_data(server_data)

    def download_dependency_build_log(self, file_directory="."):
        """Get log of a custom task version dependency build.
        .. versionadded:: v2.27

        Parameters
        ----------
        file_directory: str (optional, default is ".")
            Directory path where downloaded file is to save.

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """

        url = self._dependency_build_log_path.format(self.custom_task_id, self.id)
        response = self._client.get(url)
        content_disposition = response.headers["Content-Disposition"]
        fiie_name = content_disposition[len("attachment; filename=") :]
        with open(os.path.join(file_directory, fiie_name), "wb") as f:
            for chunk in response.iter_content(chunk_size=1024 * 1024):
                f.write(chunk)



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import json

import pytest
import responses

from datarobot import Model
from datarobot.errors import ClientError
from datarobot.models.pareto_front import ParetoFront, Solution


@pytest.fixture
def pareto_front_data(project_id):
    return {
        "project_id": project_id,
        "error_metric": "Log Loss Error",
        "hyperparameters": {},
        "target_type": "Binary",
        "solutions": [
            {
                "eureqa_solution_id": "5b2940f202d5abf17ddca9f8",
                "complexity": 1,
                "error": 0.6568885703145153,
                "expression": "target = 0",
                "expression_annotated": "target = 0",
                "best_model": False,
            },
            {
                "eureqa_solution_id": "5b2940f202d5abf17ddca9f9",
                "complexity": 2,
                "error": 0.6463236186153043,
                "expression": "target = 0",
                "expression_annotated": "target = 0",
                "best_model": False,
            },
            {
                "eureqa_solution_id": "5b2940f202d5abf17ddca9fa",
                "complexity": 3,
                "error": None,
                "expression": "target = 0",
                "expression_annotated": "target = 0",
                "best_model": False,
            },
            {
                "eureqa_solution_id": "5b2940f202d5abf17ddca9fb",
                "complexity": 4,
                "error": 0.6341910672553971,
                "expression": "target = 0",
                "expression_annotated": "target = 0",
                "best_model": False,
            },
            {
                "eureqa_solution_id": "5b29406c6523cd0665685a8d",
                "complexity": 5,
                "error": 0.6327990044770845,
                "expression": "target = 0",
                "expression_annotated": "target = 0",
                "best_model": True,
            },
        ],
    }


@pytest.fixture
def server_pareto_front_data(project_id):
    return {
        "projectId": project_id,
        "errorMetric": "Log Loss Error",
        "hyperparameters": {"random stuff": True, "other random stuff": "look ma, no trafaret!"},
        "targetType": "Binary",
        "solutions": [
            {
                "eureqaSolutionId": "5b2940f202d5abf17ddca9f8",
                "complexity": 1,
                "error": 0.6568885703145153,
                "expression": "target = 0",
                "expressionAnnotated": "target = 0",
                "bestModel": False,
            },
            {
                "eureqaSolutionId": "5b2940f202d5abf17ddca9f9",
                "complexity": 2,
                "error": 0.6463236186153043,
                "expression": "target = 0",
                "expressionAnnotated": "target = 0",
                "bestModel": False,
            },
            {
                "eureqaSolutionId": "5b2940f202d5abf17ddca9fa",
                "complexity": 3,
                "error": None,
                "expression": "target = 0",
                "expressionAnnotated": "target = 0",
                "bestModel": False,
            },
            {
                "eureqaSolutionId": "5b2940f202d5abf17ddca9fb",
                "complexity": 4,
                "error": 0.6341910672553971,
                "expression": "target = 0",
                "expressionAnnotated": "target = 0",
                "bestModel": False,
            },
            {
                "eureqaSolutionId": "5b29406c6523cd0665685a8d",
                "complexity": 5,
                "error": 0.6327990044770845,
                "expression": "target = 0",
                "expressionAnnotated": "target = 0",
                "bestModel": True,
            },
        ],
    }


def test_instantiation(server_pareto_front_data, project_id):
    pareto_front = ParetoFront.from_server_data(server_pareto_front_data)

    assert pareto_front.project_id == project_id
    assert pareto_front.error_metric == "Log Loss Error"
    assert pareto_front.hyperparameters == {
        "random stuff": True,
        "other random stuff": "look ma, no trafaret!",
    }
    assert pareto_front.target_type == "Binary"


def test_future_proof(server_pareto_front_data, project_id):
    data_with_future_keys = dict(server_pareto_front_data, new_key="amazing new key")
    data_with_future_keys["solutions"][0]["new_key"] = "amazing new key"
    pareto_front = ParetoFront.from_server_data(data_with_future_keys)

    assert pareto_front.project_id == project_id
    assert pareto_front.error_metric == "Log Loss Error"
    assert pareto_front.hyperparameters == {
        "random stuff": True,
        "other random stuff": "look ma, no trafaret!",
    }
    assert pareto_front.target_type == "Binary"


@pytest.fixture
def pareto_front_data_url(project_id, model_id):
    return "https://host_name.com/projects/{}/eureqaModels/{}/".format(project_id, model_id)


@responses.activate
def test_get_pareto_front(
    server_pareto_front_data, pareto_front_data, pareto_front_data_url, project_id, model_id
):
    responses.add(
        responses.GET,
        pareto_front_data_url,
        status=200,
        content_type="application/json",
        body=json.dumps(server_pareto_front_data),
    )
    model = Model(id=model_id, project_id=project_id)
    pf = model.get_pareto_front()
    mock_data = pareto_front_data

    assert pf.project_id == project_id
    assert pf.error_metric == "Log Loss Error"
    assert pf.hyperparameters == {
        "random stuff": True,
        "other random stuff": "look ma, no trafaret!",
    }
    assert pf.target_type == "Binary"

    for i in range(0, 5):
        assert (
            pf.solutions[i].__dict__
            == Solution(project_id=project_id, **mock_data["solutions"][i]).__dict__
        )


@responses.activate
def test_non_eureqa_model(pareto_front_data_url, project_id, model_id):
    responses.add(
        responses.GET,
        pareto_front_data_url,
        status=404,
        content_type="application/json",
        body=json.dumps({}),
    )
    model = Model(id=model_id, project_id=project_id)

    with pytest.raises(ClientError):
        model.get_pareto_front()



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import pytest
import responses

from datarobot import Deployment
from tests.utils import request_body_to_json


@responses.activate
@pytest.mark.usefixtures("deployment_get_response")
def test_config_replacement(
    unittest_endpoint, deployment_data,
):
    deployment_id = deployment_data["id"]

    url = "{}/deployments/{}/model/secondaryDatasetConfiguration/".format(
        unittest_endpoint, deployment_id
    )
    responses.add(
        responses.PATCH, url, status=202, content_type="application/json",
    )
    responses.add(
        responses.GET, url, json={"secondary_dataset_config_id": "5df109112ca582033ff44084"}
    )
    deployment = Deployment.get(deployment_id)
    result = deployment.update_secondary_dataset_config("5df109112ca582033ff44084")

    request_body = request_body_to_json(responses.calls[1].request)
    assert request_body["secondaryDatasetConfigId"] == "5df109112ca582033ff44084"

    assert result == "5df109112ca582033ff44084"


@responses.activate
@pytest.mark.usefixtures("deployment_get_response")
def test_get_config(unittest_endpoint, deployment_data):
    deployment_id = deployment_data["id"]

    url = "{}/deployments/{}/model/secondaryDatasetConfiguration/".format(
        unittest_endpoint, deployment_id
    )
    responses.add(
        responses.GET, url, json={"secondary_dataset_config_id": "5df109112ca582033ff44084"}
    )

    deployment = Deployment.get(deployment_id)
    result = deployment.get_secondary_dataset_config()

    assert responses.calls[0].request.method == "GET"
    assert result == "5df109112ca582033ff44084"
